# This script is used to inspect python binary installations, it needs to work adhoc (std lib only)
import json
import platform
import sys


def simple_inspection():
    return dict(
        version=".".join(str(s) for s in sys.version_info[:3]),
        machine=platform.machine(),
    )


if __name__ == "__main__":
    if len(sys.argv) == 1:
        print(json.dumps(simple_inspection()))



import json
import os
import re
from pathlib import Path

from runez.file import ls_dir
from runez.http import RestClient, urljoin
from runez.program import is_executable, run
from runez.system import _R, cached_property, flattened, joined, ltattr, resolved_path, short, UNSET


CPYTHON = "cpython"
RX_PYTHON_BASENAME = re.compile(r"^python(\d(\.\d+)?)?$")


class ArtifactInfo:
    """Info extracted from a typical python build artifact basename"""

    def __init__(
        self,
        basename,
        package_name,
        version,
        is_wheel=False,
        source=None,
        tags=None,
        wheel_build_number=None,
        last_modified=None,
        size=None,
    ):
        """
        Args:
            basename (str): Basename of artifact
            package_name (str): Package name, may not be completely standard
            version (Version): Package version
            is_wheel (bool): Is this artifact a wheel?
            source: Optional arbitrary object to track provenance of ArtifactInfo
            tags (str | None): Wheel tags, if any
            wheel_build_number (str | None): Wheel build number, if any
            last_modified (datetime.datetime | None): Timestamp when artifact was last modified, if available
            size (int | None): Size in bytes of artifact, if available
        """
        self.basename = basename
        self.package_name = package_name
        self.version = version
        self.is_wheel = is_wheel
        self.source = source
        self.tags = tags
        self.wheel_build_number = wheel_build_number
        self.pypi_name = PypiStd.std_package_name(package_name)
        self.relative_url = "%s/%s" % (self.pypi_name, basename)
        self.last_modified = last_modified
        self.size = size

    @classmethod
    def from_basename(cls, basename, source=None, last_modified=None, size=None):
        """
        Args:
            basename (str): Basename to parse
            source: Optional arbitrary object to track provenance of ArtifactInfo
            last_modified (datetime.datetime | None): Timestamp when artifact was last modified, if available
            size (int | None): Size in bytes of artifact, if available

        Returns:
            (ArtifactInfo | None): Parsed artifact info, if any
        """
        is_wheel = wheel_build_number = tags = None
        m = PypiStd.RX_SDIST.match(basename)
        if not m:
            m = PypiStd.RX_WHEEL.match(basename)
            if not m:
                return None

            wheel_build_number = m.group(4)
            tags = m.group(5)
            is_wheel = True

        # RX_SDIST and RX_WHEEL both yield package_name and version as match groups 1 and 2
        return cls(
            basename,
            m.group(1),
            Version(m.group(2), canonical=None),
            is_wheel=is_wheel,
            source=source,
            tags=tags,
            wheel_build_number=wheel_build_number,
            last_modified=last_modified,
            size=size,
        )

    def __repr__(self):
        return self.relative_url or self.basename

    def __eq__(self, other):
        return isinstance(other, ArtifactInfo) and self.basename == other.basename

    def __lt__(self, other):
        """Ordered by source, then pypi_name, then version, then category"""
        return ltattr(self, other, "source", "pypi_name", "version", "category", t=ArtifactInfo)

    @property
    def category(self):
        return "wheel" if self.is_wheel else "sdist"

    @property
    def is_dirty(self):
        return self.version.is_dirty


class PypiStd:
    """
    Check/standardize pypi package names
    More strict than actual pypi (for example: names starting with a number are not considered value)
    """

    RX_ACCEPTABLE_PACKAGE_NAME = re.compile(r"^[a-z][\w.-]*[a-z\d]$", re.IGNORECASE)
    RR_PYPI = re.compile(r"([^a-z\d.-]+|--+)", re.IGNORECASE)
    RR_WHEEL = re.compile(r"[^a-z\d.]+", re.IGNORECASE)

    RX_HREF = re.compile(r'href=".+/([^/#]+\.(tar\.gz|whl))#', re.IGNORECASE)
    RX_SDIST = re.compile(r"^([a-z][\w.-]*[a-z\d])-(\d[\w.!+-]*)\.tar\.gz$", re.IGNORECASE)
    RX_WHEEL = re.compile(r"^([a-z][\w.]*[a-z\d])-(\d[\w.!+]*)(-(\d[\w.]*))?-(.*)\.whl$", re.IGNORECASE)

    DEFAULT_PYPI_URL = "https://pypi.org/pypi/{name}/json"
    _pypi_client = None

    @classmethod
    def is_acceptable(cls, name):
        """Is 'name' an acceptable pypi package name?"""
        return bool(isinstance(name, str) and name != "UNKNOWN" and cls.RX_ACCEPTABLE_PACKAGE_NAME.match(name))

    @classmethod
    def std_package_name(cls, name):
        """Standardized pypi package name, single dashes and alphanumeric chars allowed only"""
        if cls.is_acceptable(name):
            name = name.replace(".", "-")
            dashed = cls.RR_PYPI.sub("-", name).lower()
            return cls.RR_PYPI.sub("-", dashed)  # 2nd pass to ensure no `--` remains

    @classmethod
    def std_wheel_basename(cls, name):
        """Standardized wheel file base name, single underscores, dots and alphanumeric chars only"""
        if cls.is_acceptable(name):
            return cls.RR_WHEEL.sub("_", name)

    @classmethod
    def default_pypi_client(cls):
        """
        Returns:
            (RestClient): Default client to use to query pypi
        """
        if cls._pypi_client is None:
            cls._pypi_client = RestClient()

        return cls._pypi_client

    @classmethod
    def pypi_response(cls, package_name, client=None, index=None, fatal=False, logger=False):
        """See https://warehouse.pypa.io/api-reference/json/
        Args:
            package_name (str): Pypi package name
            client (RestClient | None): Optional custom pypi client to use
            index (str | None): Optional custom pypi index url
            fatal (type | bool | None): True: abort execution on failure, False: don't abort but log, None: don't abort, don't log
            logger (callable | bool | None): Logger to use, True to print(), False to trace(), None to disable log chatter

        Returns:
            (dict | str | None): Dict if we queried actual REST endpoint, html text otherwise (legacy pypi simple)
        """
        pypi_name = cls.std_package_name(package_name)
        if not pypi_name:
            return None

        if client:
            if not index:
                index = client.base_url

        else:
            client = cls.default_pypi_client()

        if not index:
            index = cls.DEFAULT_PYPI_URL

        if "{name}" in index:
            url = index.format(name=pypi_name)

        else:
            url = urljoin(index, "%s/" % pypi_name)

        r = client.get_response(url, fatal=fatal, logger=logger)
        if r and r.ok:
            text = (r.text or "").strip()
            if text.startswith("{"):
                return json.loads(text)

            return text

    @classmethod
    def latest_pypi_version(cls, package_name, client=None, index=None, include_prerelease=False, fatal=False, logger=False):
        """
        Args:
            package_name (str): Pypi package name
            client (RestClient | None): Optional custom pypi client to use
            index (str | None): Optional custom pypi index url
            include_prerelease (bool): If True, include pre releases
            fatal (type | bool | None): True: abort execution on failure, False: don't abort but log, None: don't abort, don't log
            logger (callable | bool | None): Logger to use, True to print(), False to trace(), None to disable log chatter

        Returns:
            (Version | None): Latest version, if any
        """
        response = cls.pypi_response(package_name, client=client, index=index, fatal=fatal, logger=logger)
        if response:
            if isinstance(response, dict):
                info = response.get("info")
                if isinstance(info, dict) and not info.get("yanked"):  # Not sure if this can ever happen
                    version = Version(info.get("version"))
                    if version.is_valid and (include_prerelease or not version.prerelease):
                        return version

                versions = sorted(x.version for x in cls._versions_from_pypi(response.get("releases")))

            else:
                versions = sorted(i.version for i in cls._parsed_legacy_html(response))

            if not include_prerelease:
                candidates = [v for v in versions if v.is_valid and not v.prerelease]
                if candidates:
                    versions = candidates

            if versions:
                return versions[-1]

    @classmethod
    def ls_pypi(cls, package_name, client=None, index=None, source=None, fatal=False, logger=False):
        """
        Args:
            package_name (str): Pypi package name
            client (RestClient | None): Optional custom pypi client to use
            index (str | None): Optional custom pypi index url
            source: Optional arbitrary object to track provenance of ArtifactInfo
            fatal (type | bool | None): True: abort execution on failure, False: don't abort but log, None: don't abort, don't log
            logger (callable | bool | None): Logger to use, True to print(), False to trace(), None to disable log chatter

        Returns:
            (list[ArtifactInfo] | None): Artifacts reported by pypi mirror
        """
        response = cls.pypi_response(package_name, client=client, index=index, fatal=fatal, logger=logger)
        if isinstance(response, dict):
            yield from cls._versions_from_pypi(response.get("releases"), source=source)

        elif response:
            yield from cls._parsed_legacy_html(response, source=source)

    @classmethod
    def _versions_from_pypi(cls, releases, source=None):
        if isinstance(releases, dict):
            for v, infos in releases.items():
                for info in infos:
                    if not info.get("yanked"):
                        size = info.get("size")
                        upload_time = _R.lc.rm.to_datetime(info.get("upload_time"))
                        info = ArtifactInfo.from_basename(info.get("filename"), last_modified=upload_time, source=source, size=size)
                        if info:
                            yield info

    @classmethod
    def _parsed_legacy_html(cls, text, source=None):
        """
        Args:
            text (str): Text as received from a legacy pypi server
            source: Optional arbitrary object to track provenance of ArtifactInfo

        Yields:
            (ArtifactInfo): Extracted information
        """
        if text:
            lines = text.strip().splitlines()
            if lines and "does not exist" not in lines[0]:
                for line in lines:
                    m = cls.RX_HREF.search(line)
                    if m:
                        info = ArtifactInfo.from_basename(m.group(1), source=source)
                        if info:
                            yield info


class PythonSpec:
    """
    Internal canonical reference to a desired python installation, used to find python installations in `PythonDepot`
    Examples: cpython:3, cpython:3.9, pypy:3.9
    """

    def __init__(self, family, version, is_min_spec=False):
        """
        Args:
            family (str): Python family (cpython, conda, pypi, ...)
            version (Version | str): Desired version
            is_min_spec (bool): If True, match installations that are at least at `version`, e.g. cpython:3.10+
        """
        self.family = family
        self.version = Version.from_object(version)
        self.canonical = "%s:%s%s" % (family, version, "+" if is_min_spec else "")
        self.is_min_spec = is_min_spec

    def __repr__(self):
        return self.canonical

    def __hash__(self):
        return hash(self.canonical)

    def __eq__(self, other):
        return isinstance(other, PythonSpec) and self.canonical == other.canonical

    def __lt__(self, other):
        return ltattr(self, other, "family", "version", t=PythonSpec)

    def satisfies(self, other):
        """Does this spec satisfy 'other'?"""
        if isinstance(other, PythonSpec) and self.family == other.family:
            if other.is_min_spec:
                return self.version >= other.version

            return self.canonical.startswith(other.canonical)

    def represented(self, color=None, compact=CPYTHON):
        """
        Args:
            color (callable | None): Optional color to use
            compact (str | list | set | tuple | bool | None): Show version only, if `self.family` is mentioned in `compact`

        Returns:
            (str): Textual representation of this spec
        """
        text = self.canonical
        if compact and (compact is True or self.family in compact):
            text = self.version.text
            if self.is_min_spec:
                text += "+"

        return _R.colored(text, color)

    @classmethod
    def from_text(cls, text):
        """
        Args:
            text (str | None): Text to be converted into a PythonSpec() if possible, eg: 3.10, py310, python3.10, conda:3.10

        Returns:
            (PythonSpec | None): Parsed spec from given object, if valid
        """
        m = re.match(r"^(py|python|)(?P<version>\d+(\.\d+(.\w+)*)?)?(?P<min_spec>\+?)$", text)
        if m:
            version = Version.from_tox_like(m.group("version"), default="3")
            return cls(CPYTHON, version, is_min_spec=bool(m.group("min_spec"))) if version else None

        m = re.match(r"^(?P<family>cpython|conda|pypy):(?P<version>\d.*)$", text)
        if m:
            min_spec = False
            version = m.group("version")
            if version.endswith("+"):
                min_spec = True
                version = version[:-1]

            version = Version.from_tox_like(version)
            if version and version.is_valid:
                return cls(m.group("family"), version, is_min_spec=min_spec)

    @classmethod
    def from_object(cls, value):
        """
        Args:
            value: Value to transform into a PythonSpec, if possible

        Returns:
            (PythonSpec | None): Parsed spec from given object, if valid
        """
        if not value or isinstance(value, PythonSpec):
            return value or None

        if isinstance(value, Version):
            return cls(CPYTHON, value)

        if value:
            return cls.from_text(str(value))

    @classmethod
    def to_list(cls, values):
        """
        Args:
            values: Values to transform into a list of PythonSpec-s

        Returns:
            (list[PythonSpec]): Corresponding list of PythonSpec-s
        """
        values = flattened(values, split=",", transform=PythonSpec.from_object)
        values = [x for x in values if x and x.version]
        return values

    @classmethod
    def guess_family(cls, text):
        """
        Args:
            text (str | None): Text to examine

        Returns:
            (str): Guessed python family from given 'text' (typically path to installation)
        """
        if text:
            if "forge" in text or "conda" in text:
                return "conda"

            if "pypy" in text:
                return "pypy"

        return CPYTHON


class PythonDepot:
    """
    Find python installations on current machine

    Example usage:
        my_depot = PythonDepot("~/.pyenv/versions/**", "PATH")
        p = my_depot.find_python("3.7")
    """

    _preferred_python = None  # type: PythonInstallation  # Preferred python to use, if configured

    def __init__(self, *locations):
        """
        Args:
            locations (str | Path): Locations to scan
        """
        self.locations = [PythonInstallationLocation.from_location(x) for x in flattened(locations)]
        self.invoker = _R.lc.rm.SYS_INFO.invoker_python

    @cached_property
    def available_pythons(self):
        result = []
        for location in self.locations:
            result.extend(location.available_pythons)

        return result

    @property
    def preferred_python(self):
        """
        Returns:
            (PythonInstallation | None): Preferred python to use, either explicitly configured, or auto-determined from locations
        """
        if self._preferred_python is None:
            for location in self.locations:
                python = location.preferred_python
                if python:
                    return python

        return self._preferred_python

    def set_preferred_python(self, *specs):
        """
        Args:
            specs (str | list[str]): List of preferred pythons, eg: 3.10, /usr/bin/python3, invoker
        """
        for spec in flattened(specs):
            python = self.find_python(spec)
            if python and not python.problem:
                self._preferred_python = python
                return

    def find_python(self, spec):
        """
        Args:
            spec (str | Path | PythonInstallation | PythonSpec | Version | None): Example: 3.7, py37, pypy:3.7, /usr/bin/python3

        Returns:
            (PythonInstallation): Object representing python installation (may not be usable, see reported .problem)
        """
        if not spec:
            return self.preferred_python or self.invoker

        if isinstance(spec, PythonInstallation):
            return spec

        if isinstance(spec, Path):
            return PythonInstallation.from_path(spec, short_name=short(spec))

        python = self._find_python(spec)
        if python is None:
            if not isinstance(spec, Path):
                spec = Path(str(spec))

            python = PythonInstallation(spec)

        return python

    def _find_python(self, spec):
        if isinstance(spec, str):
            if spec.startswith("~") or spec.startswith(".") or "/" in spec or os.path.exists(spec):
                return PythonInstallation.from_path(Path(resolved_path(spec)), short_name=short(spec))

            elif spec == "invoker":
                return self.invoker

            spec = PythonSpec.from_text(spec)

        else:
            spec = PythonSpec.from_object(spec)

        if not isinstance(spec, PythonSpec):
            return None

        preferred = self.preferred_python
        if preferred and preferred.satisfies(spec):
            return preferred

        for location in self.locations:
            python = location.find_python(spec)
            if python is not None:
                return python

        if self.invoker and self.invoker.satisfies(spec):
            return self.invoker

    def representation(self):
        """(str): Textual representation of available pythons"""
        result = []
        for location in self.locations:
            result.append(location.representation())

        return joined(result, delimiter="\n\n") or "No PythonDepot locations configured"


class Version:
    """
    Parse versions according to PEP-440, including ordering.
    """

    text: str

    def __init__(self, text, max_parts=5, canonical=False):
        """
        Args:
            text (str | None): Text to be parsed
            max_parts (int): Maximum number of parts (components) to consider version valid
            canonical (bool | None): None: loose parsing, False: strict parsing, version left as-is, True: Turn into canonical PEP-440
        """
        self.given_text = text
        self.given_components = None  # Components as given by 'text'
        self.text = text or ""
        self.components = None  # tuple of components with exactly 'max_parts', autofilled with zeros
        self.epoch = 0
        self.local_part = None
        self.prerelease = None
        self.suffix = None
        m = _R.lc.rx_version.match(self.text)
        if not m:
            self.ignored = self.text
            return

        self.ignored = m.group("ignored") or None
        # 'canonical' set to None allows to continue even if there were extraneous bits
        if canonical is not None and self.ignored:
            return

        self.text = m.group("vtext")
        self.epoch = int(m.group("epoch") or 0)
        self.local_part = m.group("local") or None
        pre, pre_num, rel, rel_num, dev, dev_num = m.group("pre", "pre_num", "rel", "rel_num", "dev", "dev_num")
        if pre == "c":
            pre = "rc"

        # Order: .devN, aN, bN, rcN, <no suffix>, .postN
        self.suffix = joined(pre, rel, dev, delimiter=".", keep_empty=None) or None
        if pre or dev:
            self.prerelease = pre or "", int(pre_num or 0), rel or "", int(rel_num or 0), dev or "z", int(dev_num or 0)
            if pre:
                rel = rel_num = None  # rc.post does not count as .post (but a .post.dev does)

        components = [int(c) for c in m.group("main").split(".")]
        if len(components) > max_parts:
            return  # Invalid version, too many parts

        self.given_components = tuple(components)
        while len(components) < max_parts:
            components.append(0)

        self.release_number = None if rel_num is None else int(rel_num or 0)
        components.append(int(rel_num or 0))
        components.append(rel or "")
        self.components = tuple(components)
        if canonical is True:
            self.text = self.pep_440

    @classmethod
    def extracted_from_text(cls, text):
        """
        Args:
            text (str): Text to extract version from

        Returns:
            (Version | None): Parsed version, if a valid one was found
        """
        m = _R.lc.rx_version.search(text)
        if m:
            v = cls(m.group("vtext"), canonical=None)
            if v.is_valid:
                return v

    @classmethod
    def from_object(cls, obj):
        """
        Args:
            obj: Object to turn into a Version, if possible

        Returns:
            (Version | None): Corresponding version object, if valid
        """
        if obj:
            if isinstance(obj, Version):
                return obj if obj.is_valid else None

            if not isinstance(obj, str):
                obj = joined(obj, delimiter=".")

            v = cls(obj)
            if v.is_valid:
                return v

    @classmethod
    def from_tox_like(cls, text, default=None):
        """Parse a version taking into account tox-like versions like '310' meaning '3.10'"""
        if text and re.match(r"^(\d\d+)$", text):
            return cls.from_object((text[0], text[1:]))

        return cls.from_object(text or default)

    def __repr__(self):
        return self.text

    def __hash__(self):
        return hash(self.text)

    def __eq__(self, other):
        if not self.is_valid:
            return isinstance(other, Version) and self.text == other.text

        other = Version.from_object(other)
        return (
            isinstance(other, Version)
            and self.epoch == other.epoch
            and self.components == other.components
            and self.prerelease == other.prerelease
            and self.local_part == other.local_part
        )

    def __lt__(self, other):
        other = Version.from_object(other)
        if isinstance(other, Version):
            if self.epoch != other.epoch:
                return self.epoch < other.epoch

            if self.components is None or other.components is None:
                return bool(other.components)

            if self.components == other.components:
                if self.prerelease == other.prerelease:
                    my_parts = self.local_parts
                    other_parts = other.local_parts
                    if my_parts is None or other_parts is None:
                        return bool(other_parts)

                    for mine, theirs in zip(my_parts, other_parts):
                        if mine == theirs:
                            continue

                        if theirs.isdigit():
                            return not mine.isdigit() or int(mine) < int(theirs)

                        if mine.isdigit():
                            return False

                        return mine < theirs

                    return len(my_parts) < len(other_parts)

                if self.prerelease is None or other.prerelease is None:
                    return bool(self.prerelease)

                return self.prerelease < other.prerelease

            return self.components < other.components

    def __le__(self, other):
        other = Version.from_object(other)
        return isinstance(other, Version) and self < other or self == other

    def __ge__(self, other):
        other = Version.from_object(other)
        return other is None or other <= self

    def __gt__(self, other):
        other = Version.from_object(other)
        return other is None or other < self

    @cached_property
    def given_components_count(self):
        return len(self.given_components) if self.given_components else 0

    @cached_property
    def local_parts(self):
        """Local parts are only needed when comparing versions that differ solely by local part..."""
        if self.local_part:
            v = getattr(self, "_local_parts", None)
            if v is None:
                v = self.local_part.split(".")
                setattr(self, "_local_parts", v)

            return v

    @cached_property
    def is_dirty(self):
        """Is this version marked as 'dirty'?"""
        return self.local_part and "dirty" in self.local_part

    @cached_property
    def is_final(self):
        """Is this a final version as per PEP-440?"""
        return self.is_valid and not self.prerelease

    @cached_property
    def is_valid(self):
        """Is this version valid?"""
        return self.components is not None

    @cached_property
    def main(self):
        """(str): Main part of version (Major.minor.patch)"""
        if self.given_components:
            return ".".join(str(x) for x in self.given_components[:3])

    @cached_property
    def major(self):
        """(int): Major part of version"""
        if self.given_components_count >= 1:
            return self.components and self.components[0]

    @cached_property
    def minor(self):
        """(int): Minor part of version"""
        if self.given_components_count >= 2:
            return self.components and self.components[1]

    @cached_property
    def mm(self):
        """(str): <major>.<minor>, often used in python paths, like config-3.9"""
        if self.minor is not None:
            return "%s.%s" % (self.major, self.minor)

    @property
    def patch(self):
        """(int): Patch part of version"""
        if self.given_components_count >= 3:
            return self.components and self.components[2]

    @cached_property
    def pep_440(self):
        """PEP-440 canonical version"""
        if self.is_valid:
            result = []
            if self.epoch:
                result.append(f"{self.epoch}!")

            result.append(self.main)
            if self.prerelease:
                if self.prerelease[0]:
                    result.append("%s%s" % (self.prerelease[0], self.prerelease[1]))

                if self.prerelease[2]:
                    result.append(".%s%s" % (self.prerelease[2], self.prerelease[3]))

                if self.prerelease[4] and self.prerelease[4] != "z":
                    result.append(".%s%s" % (self.prerelease[4], self.prerelease[5]))

            if self.release_number is not None:
                result.append(".post%s" % self.release_number)

            if self.local_part:
                result.append(f"+{self.local_part}")

            return "".join(result)


class PythonInstallation:
    """Models a specific python installation"""

    def __init__(self, executable, short_name=None):
        """
        Args:
            executable (Path): Path to python executable
            short_name (str | None): Short name to textually represent this installation
        """
        self.executable = executable
        self.short_name = short_name or short(executable)
        self.real_exe, inspection = PythonSimpleInspection.exe_inspection(executable)
        self.inspection = inspection
        self.problem = inspection.problem
        version = None
        if not self.problem:
            version = Version(inspection.version)
            if not version.is_valid:
                self.problem = f"invalid version '{version.text}'"
                version = None

        self.full_version = version

    def __repr__(self):
        return self.representation(is_coloring=False)

    def __str__(self):
        return self.representation()

    def __eq__(self, other):
        return isinstance(other, PythonInstallation) and self.real_exe == other.real_exe

    def __lt__(self, other):
        if isinstance(other, PythonInstallation):
            if self.full_spec is None or other.full_spec is None:
                return other.full_spec is not None or str(self.executable) < str(other.executable)

            if self.full_spec == other.full_spec:
                return str(self.executable) < str(other.executable)

            return self.full_spec < other.full_spec

    def representation(self, delimiter=", ", is_coloring=UNSET, more_info=None):
        """
        Args:
            delimiter (str): Delimiter to use in the more info section
            is_coloring (bool): Whether text should be colored (default: yes on tty)
            more_info (list[str] | None): Optional list of strings to append to textual representation

        Returns:
            (str): Short textual representation of this installation
        """
        if more_info is None:
            more_info = []

        if self.machine and self.machine != _R.lc.rm.SYS_INFO.platform_id.arch:
            more_info.append(_R.colored(self.machine, "dim", is_coloring=is_coloring))

        if self.is_invoker:
            more_info.append(_R.colored("invoker", "green", is_coloring=is_coloring))

        full_spec = self.full_spec
        if full_spec and full_spec.version.text not in self.short_name:
            more_info.append(full_spec.represented())

        if self.problem:
            more_info.append(_R.colored(self.problem, "red", is_coloring=is_coloring))

        text = _R.colored(self.short_name, "bold", is_coloring=is_coloring)
        if more_info:
            return "%s [%s]" % (text, delimiter.join(reversed(more_info)))

        return text

    @classmethod
    def from_path(cls, path, short_name=None):
        """
        Args:
            path (Path): Path to folder or file
            short_name (str | None): Short name to use for textual representation of this installation

        Returns:
            (PythonInstallation): Associated installation
        """
        if path.is_dir():
            return cls.from_folder(path, short_name=short_name)

        return cls(path, short_name=short_name)

    @classmethod
    def from_folder(cls, folder, short_name=None):
        """
        Args:
            folder (Path): Path to folder
            short_name (str | None): Short name to use for textual representation of this installation

        Returns:
            (PythonInstallation): Associated installation
        """
        if folder.name != "bin":
            bin_folder = folder / "bin"
            if bin_folder.is_dir():
                folder = bin_folder

        path = folder / "python3"
        if not is_executable(path):
            path = folder / "python"

        return cls(path, short_name=short_name)

    @property
    def is_invoker(self):
        """Is this the python installation current execution is using?"""
        return self == _R.lc.rm.SYS_INFO.invoker_python

    @cached_property
    def family(self):
        return PythonSpec.guess_family(str(self.real_exe))

    @cached_property
    def full_spec(self):
        """Full spec, , e.g: cpython:3.11.6"""
        if self.full_version:
            return PythonSpec(self.family, self.full_version)

    @cached_property
    def mm(self):
        """Major/minor version, e.g: cpython:3.11"""
        if self.full_version:
            return Version(self.full_version.mm)

    @cached_property
    def mm_spec(self):
        """Major/minor spec, e.g: cpython:3.11"""
        if self.mm:
            return PythonSpec(self.family, self.mm)

    @cached_property
    def machine(self):
        return self.inspection.machine

    def satisfies(self, given_spec):
        """
        Args:
            given_spec (PythonSpec): Spec expressed by user or configuration

        Returns:
            (bool): True if this python installation satisfies it
        """
        if given_spec and self.family == given_spec.family and not self.problem:
            spec = self.full_spec if given_spec.version.given_components_count > 2 else self.mm_spec
            return spec.satisfies(given_spec)


class PythonInstallationLocation:
    """
    Local source where python installations can be found

    Locations can be specified as:
    - <some-folder>: look for <some-folder>/pythonM.m exes
    - <some-folder>/python*: look for <some-folder>/pythonM.m/bin/pythonM.m (eg: /apps/python*)
    - <some-folder>/**: look for <some-folder>/**/bin/pythonM.m (eg: ~/.pyenv/versions/**)
    - PATH: Scan PATH env var for python exes
    """
    def __init__(self, location):
        self.location = location
        self._preferred_python = UNSET  # type: PythonInstallation # Auto-selected preferred python from this location

    def __repr__(self):
        return short(self.location)

    @classmethod
    def from_location(cls, location):
        if location == "PATH":
            return PythonInstallationLocationPathEnvVar(location)

        if location.endswith("/**"):
            return PythonInstallationLocationPyenv(location)

        if location.endswith("/python*"):
            return PythonInstallationLocationSubFolders(location)

        return PythonInstallationLocation(location)

    def _record_preferred(self, executable: Path):
        """Record 'executable' as the automatically selected preferred python for this location"""
        if not self._preferred_python and executable.is_symlink():
            symlink = os.readlink(executable)
            m = RX_PYTHON_BASENAME.match(symlink)
            if m and m.group(2):
                # We have a python -> pythonM.m symlink
                python = PythonInstallation(executable.parent / symlink)
                if not python.problem:
                    self._preferred_python = python

    def _auto_determined_preferred(self):
        """By default, prefer latest python with a patch version > .5"""
        for python in self.available_pythons:
            if python.full_version.patch > 5:
                return python

    def _scanned_location(self):
        """
        Returns:
            (list[PythonInstallation]): Python installations found in this location
        """
        # Default: pythons from a folder containing pythonM.m symlinks, eg: /opt/my-python-binaries/
        result = []
        for item in ls_dir(self.location):
            if item.is_file():
                m = RX_PYTHON_BASENAME.match(item.name)
                if m:
                    if m.group(2):
                        python = PythonInstallation(item)
                        if not python.problem:
                            result.append(python)

                    else:
                        self._record_preferred(item)

        return sorted(result, reverse=True)

    @property
    def preferred_python(self):
        """(PythonInstallation | None): Preferred python found in this location"""
        if self._preferred_python is UNSET:
            self._preferred_python = self._auto_determined_preferred()

        return self._preferred_python

    @cached_property
    def available_pythons(self):
        """
        Returns:
            (list[PythonInstallation]): Python installations found in this location
        """
        return self._scanned_location()

    def find_python(self, spec):
        for python in self.available_pythons:
            if python.satisfies(spec):
                return python

    def representation(self):
        """(str): Textual representation of available pythons"""
        if not self.available_pythons:
            return "No python installations found in '%s'" % _R.colored(self, "orange")

        preferred = self.preferred_python
        representation = ["%s in %s:" % (_R.lc.rm.plural(self.available_pythons, "python installation"), self)]
        for python in self.available_pythons:
            more_info = None
            if preferred and python.executable == preferred.executable:
                more_info = [_R.colored("preferred", "dim")]

            representation.append(python.representation(more_info=more_info))

        return joined(representation, delimiter="\n")


class PythonInstallationLocationPathEnvVar(PythonInstallationLocation):
    """Pythons from PATH env var"""

    def _auto_determined_preferred(self):
        """When using PATH env var, don't pick any preferred python"""
        return None

    def _scanned_location(self):
        result = []
        venv = os.environ.get("VIRTUAL_ENV")
        for folder in flattened(os.environ.get("PATH"), split=os.pathsep):
            if venv and folder.startswith(venv):
                continue  # Ignore python installations from virtualenv

            general = []  # General symlinks, eg: `python3` and `python`
            major_minors = []  # Major.minor symlinks, eg: `python3.7`
            for item in ls_dir(folder):
                if is_executable(item):
                    m = RX_PYTHON_BASENAME.match(item.name)
                    if m:
                        python = PythonInstallation(item)
                        if not python.problem:
                            target = major_minors if m.group(2) else general
                            target.append(python)

            result.extend(sorted(general, key=lambda x: x.executable))
            result.extend(sorted(major_minors, reverse=True))

        return result


class PythonInstallationLocationPyenv(PythonInstallationLocation):
    """Pythons from pyenv-like location, eg: ~/.pyenv/versions/**"""

    def _scanned_location(self):
        result = []
        for item in ls_dir(os.path.dirname(self.location)):
            if not item.is_symlink() and item.is_dir():
                bin_folder = item / "bin"
                if bin_folder.is_dir():
                    python = PythonInstallation.from_folder(bin_folder, short_name=short(item))
                    if not python.problem:
                        result.append(python)

        return sorted(result, reverse=True)


class PythonInstallationLocationSubFolders(PythonInstallationLocation):
    """Pythons from python* directories, eg: /apps/python*"""

    def _scanned_location(self):
        result = []
        for item in ls_dir(os.path.dirname(self.location)):
            if item.is_dir():
                m = RX_PYTHON_BASENAME.match(item.name)
                if m and m.group(2):
                    python = PythonInstallation(item / "bin" / item.name, short_name=short(item))
                    if not python.problem:
                        result.append(python)

        return sorted(result, reverse=True)


class PythonSimpleInspection:
    """Simple inspection (version and arch) of a python executable, cached to avoid expensive python process invocations"""

    _cached = {}

    def __init__(self, version=None, machine=None, problem=None):
        self.version = version
        self.machine = machine
        self.problem = problem

    def __repr__(self):
        return self.problem or f"{self.version} ({self.machine})"

    @classmethod
    def register(cls, key, real_exe, inspection):
        cls._cached[key] = (real_exe, inspection)
        cls._cached[real_exe] = (real_exe, inspection)

    @classmethod
    def exe_inspection(cls, executable):
        cached = cls._cached.get(executable)
        if cached is not None:
            return cached

        real_exe = executable.resolve()
        cached = cls._cached.get(real_exe)
        if cached is not None:
            return cached

        if not is_executable(executable):
            if _R.is_dryrun():
                version = Version.extracted_from_text(short(executable))
                if version and version.is_valid:
                    return real_exe, cls(version=version.text, machine=_R.lc.rm.SYS_INFO.platform_id.arch)

            return real_exe, cls(problem="not available")

        try:
            import runez._inspect

            r = run(real_exe, runez._inspect.__file__, dryrun=False, fatal=False, logger=None)
            if r.succeeded:
                result = json.loads(r.output)
                if not isinstance(result, dict) or not result.get("version"):
                    return real_exe, cls(problem=f"internal error: _inspect.py returned '{short(result)}'")

                result = cls(**result)

            else:
                result = cls(problem=short(r.full_output))

        except Exception as e:
            result = cls(problem=f"internal error: {e}")

        cls.register(executable, real_exe, result)
        return real_exe, result



import json
import os
import sys

import pytest

import runez
from runez.http import RestClient
from runez.pyenv import ArtifactInfo, PypiStd, PythonDepot, PythonSpec, Version


PYPI_CLIENT = RestClient("https://example.com/pypi")


def test_artifact_info():
    info = ArtifactInfo.from_basename("E.S.P.-Hadouken-0.2.1.tar.gz")
    assert str(info) == "e-s-p-hadouken/E.S.P.-Hadouken-0.2.1.tar.gz"
    assert info.category == "sdist"
    assert not info.is_dirty
    assert not info.is_wheel
    assert info.package_name == "E.S.P.-Hadouken"
    assert info.pypi_name == "e-s-p-hadouken"
    assert info.tags is None
    assert info.version == "0.2.1"
    assert info.wheel_build_number is None

    info = ArtifactInfo.from_basename("a_b-1!0.0.1-10-py3-none-any.whl", None)
    assert str(info) == "a-b/a_b-1!0.0.1-10-py3-none-any.whl"
    assert info.category == "wheel"
    assert not info.is_dirty
    assert info.is_wheel
    assert info.package_name == "a_b"
    assert info.pypi_name == "a-b"
    assert info.tags == "py3-none-any"
    assert info.version == "1!0.0.1"
    assert info.wheel_build_number == "10"


def mk_python(basename, executable=True, content=None, machine=None):
    if basename[0].isdigit():
        version = Version(basename)
        folder = runez.to_path(".pyenv/versions") / basename / "bin"

    else:
        version = Version(os.path.basename(basename))
        if basename.startswith("./"):
            folder = runez.to_path(os.path.dirname(basename[2:]))

        else:
            folder = runez.to_path(".pyenv/versions") / os.path.dirname(basename) / "bin"

    path = folder / ("python%s" % version.mm)
    if not content:
        content = dict(version=str(version), machine=machine or runez.SYS_INFO.platform_id.arch)

    if content == "failed":
        content = "echo failed\nexit 1"

    elif content == "invalid-json":
        content = "echo '{foo: bar}'"

    else:
        content = "echo '%s'" % json.dumps(content)

    content = "#!/bin/bash\n%s\n" % content
    runez.write(path, content, logger=None)
    runez.symlink(path, folder / "python", logger=None)
    if executable:
        runez.make_executable(path, logger=None)


def test_depot(temp_folder, logged):
    # Create some pyenv-style python installation mocks (using version 8, so it sorts above any real version...)
    mk_python("8.6.1", machine="x86_64_test")
    mk_python("8.7.2")
    mk_python("miniforge3-22.11.1-4/9.11.2")
    mk_python("pypy-9.8.7/9.8.7")
    mk_python("8.5.4", content="invalid content")
    mk_python("8.5.5", content=dict(version="invalid-version"))
    mk_python("8.5.6", content="failed")
    mk_python("8.5.7", content="invalid-json")
    runez.symlink(".pyenv/versions/8.6.1", ".pyenv/versions/8.6", logger=None)
    runez.symlink(".pyenv/versions/8.6.1", ".pyenv/versions/python8.6", logger=None)

    depot = PythonDepot(".pyenv/versions/**")
    text = depot.representation()
    assert text.startswith("4 python installations in .pyenv/versions/**:")
    assert len(depot.available_pythons) == 4

    assert str(depot.find_python("8.5.4")) == "8.5.4 [not available]"
    invalid = depot.find_python(".pyenv/versions/8.5.4")
    assert str(invalid) == ".pyenv/versions/8.5.4 [internal error: _inspect.py returned 'invalid content']"
    bad_version = depot.find_python(".pyenv/versions/8.5.5")
    assert str(bad_version) == ".pyenv/versions/8.5.5 [invalid version 'invalid-version']"
    failed = depot.find_python(".pyenv/versions/8.5.6")
    assert str(failed) == ".pyenv/versions/8.5.6 [failed]"
    invalid_json = depot.find_python(".pyenv/versions/8.5.7")
    assert str(invalid_json).startswith(".pyenv/versions/8.5.7 [internal error: ")
    assert invalid < bad_version < failed

    versions = [p.mm_spec.canonical for p in depot.available_pythons]
    assert versions == ["pypy:9.8", "cpython:8.7", "cpython:8.6", "conda:9.11"]

    # Verify that latest available is found (when under-specified)
    p8 = depot.find_python("8")
    assert repr(p8) == ".pyenv/versions/8.7.2"
    assert str(p8) == ".pyenv/versions/8.7.2"
    assert invalid < p8

    # Verify that preferred python is respected
    preferred = depot.preferred_python
    assert str(preferred) == ".pyenv/versions/pypy-9.8.7"

    depot.set_preferred_python("8.6")
    preferred = depot.preferred_python
    assert str(preferred) == ".pyenv/versions/8.6.1 [x86_64_test]"

    assert repr(depot.find_python("8")) == ".pyenv/versions/8.6.1 [x86_64_test]"
    assert p8 is depot.find_python("8.7")
    p8b = depot.find_python(".pyenv/versions/8.7.2")
    assert p8 is not p8b
    assert p8 == p8b
    assert p8 < p8b  # Due to their .path

    # Verify min spec
    assert str(depot.find_python("conda:9.1+")) == ".pyenv/versions/miniforge3-22.11.1-4 [conda:9.11.2]"

    # There's only one symlink of the form 'python*'
    depot2 = PythonDepot(".pyenv/versions/python*")
    assert len(depot2.available_pythons) == 1
    assert str(depot2.available_pythons[0]) == ".pyenv/versions/python8.6 [8.6.1, x86_64_test]"

    assert not logged


def test_depot_adhoc(temp_folder):
    depot = PythonDepot(".pyenv/versions")
    assert depot.representation() == "No python installations found in '.pyenv/versions'"

    p11 = depot.find_python("11.0.0")
    assert p11.problem == "not available"
    assert str(p11.inspection) == "not available"

    # Only paths are cached
    p11b = depot.find_python("11.0.0")
    assert p11 is not p11b
    assert p11 == p11b
    pfoo = depot.find_python("/foo")
    assert depot.find_python("/foo") == pfoo

    mk_python("./some-path/bin/13.1.2")
    assert str(depot.find_python("some-path/bin/python13.1")) == "some-path/bin/python13.1 [13.1.2]"
    assert str(depot.find_python("./some-path/")) == "./some-path/ [13.1.2]"
    assert str(depot.find_python(runez.to_path("some-path"))) == "some-path [13.1.2]"
    assert str(depot.find_python("some-path/bin/python")) == "some-path/bin/python [13.1.2]"
    assert str(depot.find_python("some-path/bin")) == "some-path/bin [13.1.2]"


def test_depot_folder(temp_folder):
    mk_python("8.5.6")
    mk_python("8.5.7")
    runez.symlink(".pyenv/versions/8.5.6/bin/python", "python8.5", logger=None)
    runez.symlink("python8.5", "python", logger=None)
    depot = PythonDepot(".")
    assert len(depot.available_pythons) == 1
    python = depot.find_python(None)
    assert str(python) == "python8.5 [8.5.6]"
    assert depot.preferred_python is python
    assert depot.locations[0].preferred_python is python
    assert depot.find_python("8.5") is python
    assert depot.find_python("8.5.6") is python
    assert str(depot.find_python("8.5.7")) == "8.5.7 [not available]"


def test_depot_path(temp_folder):
    depot = PythonDepot("PATH")
    assert depot.available_pythons
    assert depot.preferred_python is None
    assert depot.find_python(None) is depot.invoker


def test_empty_depot(temp_folder):
    depot = PythonDepot()
    assert depot.representation() == "No PythonDepot locations configured"
    assert not depot.available_pythons
    invoker = runez.SYS_INFO.invoker_python
    assert "invoker" in repr(invoker)
    assert "invoker" in str(invoker)

    p95_spec = PythonSpec.from_text("9.5")
    with runez.CaptureOutput(dryrun=True) as logged:
        # In dryrun mode, any requested python is considered available
        p95 = depot.find_python(p95_spec)
        assert str(p95) == "cpython:9.5"
        assert p95.problem is None
        assert not logged

    p95 = depot.find_python(p95_spec)
    assert p95.problem
    assert depot.find_python(None) is invoker
    assert depot.find_python("") is invoker
    assert depot.find_python("invoker") is invoker
    assert depot.find_python(invoker) is invoker
    assert depot.find_python(invoker.executable).is_invoker
    assert depot.find_python(str(invoker.executable)).is_invoker
    assert depot.find_python(invoker.real_exe).is_invoker
    assert depot.find_python(PythonSpec("cpython", invoker.mm)) is invoker
    assert depot.find_python("python") is invoker
    assert depot.find_python("py%s" % invoker.mm) is invoker  # eg: py3.10
    assert depot.find_python("py%s" % invoker.mm.text.replace(".", "")) is invoker  # tox style: py310
    assert depot.find_python(invoker.mm) is invoker
    assert depot.find_python(invoker.mm.text) is invoker
    assert depot.find_python("python") is invoker
    assert depot.find_python("python%s" % invoker.mm.major) is invoker

    p = depot.find_python("foo")
    assert str(p) == "foo [not available]"
    assert repr(p) == "foo [not available]"
    assert p.executable == runez.to_path("foo")
    assert p.real_exe == runez.to_path("foo").resolve()
    assert p.full_spec is None
    assert p.full_version is None
    assert not p.is_invoker
    assert p.machine is None
    assert p.mm_spec is None
    assert p.problem == "not available"
    assert p.short_name == "foo"


def test_inspect():
    import runez._inspect

    r = runez.run(sys.executable, runez._inspect.__file__)
    assert r.succeeded
    assert '"version":' in r.output


def test_invoker(monkeypatch):
    import runez.pyenv

    invoker = runez.SYS_INFO.invoker_python
    assert "invoker" in str(invoker)
    assert invoker.full_spec
    assert invoker.is_invoker
    assert invoker.machine
    assert invoker.problem is None


def test_pypi_standardized_naming():
    assert not PypiStd.is_acceptable(None)
    assert not PypiStd.is_acceptable(1)
    assert not PypiStd.is_acceptable("")
    assert not PypiStd.is_acceptable("a")  # Don't bother with one letter packages
    assert not PypiStd.is_acceptable("1a")  # Don't bother with anything that does start with a letter
    assert not PypiStd.is_acceptable("-a")
    assert not PypiStd.is_acceptable(".a")
    assert not PypiStd.is_acceptable("a.")  # Must end with a letter or number
    assert not PypiStd.is_acceptable("a b")  # No spaces plz

    assert PypiStd.is_acceptable("a1")
    assert PypiStd.is_acceptable("aB")
    assert PypiStd.is_acceptable("foo")
    assert PypiStd.is_acceptable("Foo_1.0")

    assert PypiStd.std_package_name(None) is None
    assert PypiStd.std_package_name(5) is None
    assert PypiStd.std_package_name("") is None
    assert PypiStd.std_package_name("-a-") is None
    assert PypiStd.std_package_name("a") is None
    assert PypiStd.std_package_name("foo") == "foo"
    assert PypiStd.std_package_name("Foo") == "foo"
    assert PypiStd.std_package_name("A__b-c_1.0") == "a-b-c-1-0"
    assert PypiStd.std_package_name("some_-Test") == "some-test"
    assert PypiStd.std_package_name("a_-_-.-_.--b") == "a-b"

    assert PypiStd.std_wheel_basename(None) is None
    assert PypiStd.std_package_name(10.1) is None
    assert PypiStd.std_wheel_basename("") is None
    assert PypiStd.std_wheel_basename("a.b_-_1.5--c") == "a.b_1.5_c"
    assert PypiStd.std_wheel_basename("a.b_-___1.5--c") == "a.b_1.5_c"


P_BLACK = """
<html><head><title>Simple Index</title><meta name="api-version" value="2" /></head><body>
<a href="/pypi/packages/pypi-public/black/black-18.3a0-py3-none-any.whl#sha256=..."</a><br/>
<a href="/pypi/packages/pypi-public/black/black-18.3a0.tar.gz#sha256=...">black-18.3a0.tar.gz</a><br/>
<a href="/pypi/packages/pypi-public/black/black-18.3a1-py3-none-any.whl#sha256=..."
"""

P_FUNKY = """
href="funky-proj/funky.proj-1.3.0+dirty_custom-py3-none-any.whl#sha256=..."
href="funky-proj/funky.proj-1.3.0_custom.tar.gz#sha256=..."
"""

P_PICKLEY = {
  "info": {"version": "2.5.6.dev1"},
  "releases": {
    "2.5.3": [{"filename": "pickley-2.5.3-py2.py3-none-any.whl", "yanked": True}, {"filename": "oops-bad-filename"}],
    "2.5.4": [{"filename": "pickley-2.5.4-py2.py3-none-any.whl", "upload_time": "2012-01-22T05:08:17"}],
    "2.5.5": [{"filename": "pickley-2.5.5-py2.py3-none-any.whl"}, {"filename": "pickley-2.5.5.tar.gz"}]
  }
}

P_SHELL_FUNCTOOLS = """
<html><head><title>Simple Index</title><meta name="api-version" value="2" /></head><body>

# 1.8.1 intentionally malformed
<a href="/pypi/shell-functools/shell_functools-1.8.1!1-py2.py3-none-any.whl9#">shell_functools-1.8.1-py2.py3-none-any.whl</a><br/>
<a href="/pypi/shell-functools/shell-functools-1.8.1!1.tar.gz#">shell-functools-1.8.1.tar.gz</a><br/>

<a href="/pypi/shell-functools/shell_functools-1.9.9+local-py2.py3-none-any.whl#sha...">shell_functools-1.9.9-py2.py3-none-any.whl</a><br/>
<a href="/pypi/shell-functools/shell-functools-1.9.9+local.tar.gz#sha256=ff...">shell-functools-1.9.9.tar.gz</a><br/>
<a href="/pypi/shell-functools/shell_functools-1.9.11-py2.py3-none-any.whl#sha256=...">shell_functools-1.9.11-py2.py3-none-any.whl</a><br/>
<a href="/pypi/shell-functools/shell-functools-1.9.11.tar.gz#sha256=ca...">shell-functools-1.9.11.tar.gz</a><br/>
</body></html>
"""


@PYPI_CLIENT.mock({
    "shell-functools/": P_SHELL_FUNCTOOLS,
    "https://pypi.org/pypi/black/json": P_BLACK,
    "https://pypi.org/pypi/foo/json": {"info": {"version": "1.2.3"}},
    "https://pypi.org/pypi/pickley/json": P_PICKLEY,
    "https://pypi.org/pypi/funky-proj/json": P_FUNKY,
})
def test_pypi_parsing():
    assert PypiStd.pypi_response("-invalid-") is None
    assert PypiStd.latest_pypi_version("shell_functools", index=PYPI_CLIENT.base_url) == Version("1.9.11")
    assert PypiStd.latest_pypi_version("foo") == Version("1.2.3")  # Vanilla case
    assert PypiStd.latest_pypi_version("pickley") == Version("2.5.5")  # Pre-release ignored
    assert PypiStd.latest_pypi_version("pickley", include_prerelease=True) == Version("2.5.6.dev1")

    assert sorted(PypiStd.ls_pypi("foo")) == []

    sample = sorted(PypiStd.ls_pypi("shell-functools", client=PYPI_CLIENT, source="s1"))
    assert len(sample) == 5
    assert str(sample[0]) == "shell-functools/shell-functools-1.8.1!1.tar.gz"
    assert sample[0].version == Version("1.8.1")
    assert not sample[0].is_dirty
    assert sample[0].category == "sdist"

    pickley = sorted(PypiStd.ls_pypi("pickley", source="s1"))
    assert len(pickley) == 3
    assert pickley[0] < sample[0]  # Alphabetical sort for same-source artifacts
    assert pickley[0].last_modified.year == 2012

    assert sample[3].version == Version("1.9.11")
    assert not sample[3].is_wheel
    assert sample[4].version == Version("1.9.11")
    assert sample[4].category == "wheel"
    assert sample[4].is_wheel
    assert sample[4].tags == "py2.py3-none-any"
    assert sample[3] < sample[4]  # Source distribution before wheel
    assert sample[3] != sample[4]

    with pytest.raises(TypeError):
        _ = sample[3] < "foo"

    black = sorted(PypiStd.ls_pypi("black"))  # All versions are pre-releases
    assert len(black) == 3
    assert black[0].version.prerelease

    funky = sorted(PypiStd.ls_pypi("funky-proj", source=None))
    assert len(funky) == 2
    assert funky[0].package_name == "funky.proj"
    assert funky[0].pypi_name == "funky-proj"
    assert funky[1].is_dirty
    assert black[0] < funky[0]  # Alphabetical sort when both have no source
    assert funky[0] < sample[4]  # Arbitrary: no-source sorts lowest...


def test_spec():
    p3 = PythonSpec.from_text("3")
    p3plus = PythonSpec.from_text("3+")
    assert str(p3) == "cpython:3"
    assert str(p3plus) == "cpython:3+"
    assert not p3.is_min_spec
    assert p3plus.is_min_spec

    assert p3 == PythonSpec.from_object(Version("3"))

    p3_rep = p3.represented()
    p3plus_rep = p3plus.represented()
    assert isinstance(p3_rep, str)
    assert isinstance(p3plus_rep, str)
    assert p3_rep == "3"
    assert p3plus_rep == "3+"

    p38 = PythonSpec.from_text("3.8")
    assert p38.satisfies(p3)
    assert p38.satisfies(p3plus)

    p38plus = PythonSpec.from_text("3.8+")
    c38 = PythonSpec.from_text("conda:3.8")
    assert c38 != p38
    assert c38.version == p38.version
    assert p38 == PythonSpec.from_text("py38")
    assert not c38.satisfies(p38)
    assert not p38.satisfies(c38)

    p310 = PythonSpec.from_text("3.10")
    assert p310 == PythonSpec.from_text("py310")
    assert not p38.satisfies(p310)
    assert not p310.satisfies(p38)
    assert p310.satisfies(p38plus)
    assert p310.satisfies(p310)
    assert PythonSpec.from_text("3.10.1").satisfies(p310)
    assert PythonSpec.from_text("3.10.0rc1").satisfies(p310)

    assert p38.represented() == "3.8"
    assert p38.represented(compact=None) == "cpython:3.8"
    assert p38.represented(compact=False) == "cpython:3.8"
    assert p38.represented(compact=True) == "3.8"
    assert p38.represented(compact="cpython") == "3.8"
    assert p38.represented(compact=["cpython", "conda"]) == "3.8"
    assert p38.represented(color=str, compact=None) == "cpython:3.8"

    assert c38.represented() == "conda:3.8"
    assert c38.represented(compact=True) == "3.8"
    assert c38.represented(compact="cpython") == "conda:3.8"
    assert c38.represented(compact=["cpython", "conda"]) == "3.8"
    assert c38.represented(color=str, compact=None) == "conda:3.8"

    rc3 = PythonSpec.from_text("cpython:3.12.0rc3")
    assert rc3.represented(compact=None) == "cpython:3.12.0rc3"
    assert not rc3.is_min_spec

    rc3plus = PythonSpec.from_text("cpython:3.12.0rc3+")
    assert rc3plus.represented(compact=True) == "3.12.0rc3+"
    assert rc3plus.is_min_spec


def test_spec_equivalent():
    def check_equivalent_specs(*text):
        specs = [PythonSpec.from_text(x) for x in text]
        assert len(specs) == len(text)
        assert len(set(specs)) == 1

    check_equivalent_specs("3", "py3", "python3", "cpython:3", "python", "py")
    check_equivalent_specs("3+", "py3+", "python3+", "cpython:3+")
    check_equivalent_specs("3.10", "310", "py3.10", "py310", "python3.10", "python310", "cpython:3.10")
    check_equivalent_specs("3.10+", "310+", "py3.10+", "py310+", "python3.10+", "python310+", "cpython:3.10+")
    check_equivalent_specs("3777", "py3.777", "python3.777", "cpython:3.777")


def test_spec_invalid():
    assert PythonSpec.from_object(None) is None
    assert PythonSpec.from_object([]) is None
    assert PythonSpec.from_object(["foo"]) is None

    def check_spec_invalid(text):
        assert PythonSpec.from_text(text) is None

    check_spec_invalid("foo")
    check_spec_invalid("foo:3")
    check_spec_invalid("cpython")
    check_spec_invalid("cpython:")
    check_spec_invalid("cpython:3 ")
    check_spec_invalid("cpython:3.10.0--rc1")
    check_spec_invalid("cpython:+")
    check_spec_invalid("cpython: 3")
    check_spec_invalid("cpython :3")
    check_spec_invalid("python:3")
    check_spec_invalid("p3.9")
    check_spec_invalid("pY3")
    check_spec_invalid("3.9.9a")
    check_spec_invalid("3.9.9++")


def test_spec_list():
    assert not PythonSpec.to_list("a,b")
    x = PythonSpec.to_list("3.10,py39")
    assert x == [PythonSpec.from_text("3.10"), PythonSpec.from_text("3.9")]

    x = PythonSpec.to_list(["3.10,py36", "foo,py37", 3.8])
    assert x == [PythonSpec.from_text("3.10"), PythonSpec.from_text("3.6"), PythonSpec.from_text("3.7"), PythonSpec.from_text("3.8")]


@pytest.mark.parametrize(
    ("given_version", "expected"),
    [
        ("1.2", (1, 2, 0, 0, 0, 0, "")),
        ("1.2rev5", (1, 2, 0, 0, 0, 5, "rev")),
        ("1.2r5.dev3", (1, 2, 0, 0, 0, 5, "r", "", 0, "r", 5, "dev", 3)),
        ("1.dev0", (1, 0, 0, 0, 0, 0, "", "", 0, "", 0, "dev", 0)),
        ("1.0.dev456", (1, 0, 0, 0, 0, 0, "", "", 0, "", 0, "dev", 456)),
        ("1.0a12", (1, 0, 0, 0, 0, 0, "", "a", 12, "", 0, "z", 0)),
        ("1.2.3rc12", (1, 2, 3, 0, 0, 0, "", "rc", 12, "", 0, "z", 0)),
        ("1.0a2.dev456", (1, 0, 0, 0, 0, 0, "", "a", 2, "", 0, "dev", 456)),
        ("1.0b2.post345", (1, 0, 0, 0, 0, 0, "", "b", 2, "post", 345, "z", 0)),
        ("1.0b2.post345.dev456", (1, 0, 0, 0, 0, 0, "", "b", 2, "post", 345, "dev", 456)),
        ("1.0rc1.dev456", (1, 0, 0, 0, 0, 0, "", "rc", 1, "", 0, "dev", 456)),
        ("1.0.post456.dev34", (1, 0, 0, 0, 0, 456, "post", "", 0, "post", 456, "dev", 34)),
    ]
)
def test_pep_sample(given_version, expected):
    version = Version(given_version)
    assert version.is_valid
    assert version.ignored is None
    assert str(version) == given_version
    actual = version.components
    if version.prerelease:
        actual += version.prerelease

    assert actual == expected


def test_version():
    loose = Version("v1.0.dirty", canonical=None)
    assert loose.is_valid
    assert loose.ignored == ".dirty"
    assert str(loose) == "1.0"
    assert loose.pep_440 == "1.0"

    invalid = Version("v1.0.dirty")
    assert not invalid.is_valid
    assert str(invalid) == "v1.0.dirty"
    assert invalid.ignored == ".dirty"
    assert loose > invalid

    dev101 = Version("0.0.1dev101")
    assert not dev101.is_final
    assert dev101.is_valid
    assert not dev101.is_dirty
    assert dev101.prerelease
    assert str(dev101) == "0.0.1dev101"
    assert dev101.pep_440 == "0.0.1.dev101"
    assert dev101.major == 0
    assert dev101.minor == 0
    assert dev101.patch == 1
    assert dev101.mm == "0.0"

    none = Version(None)
    assert str(none) == ""
    assert not none.is_valid
    assert not none.is_final
    assert none.major is None
    assert none.mm is None

    empty = Version("")
    assert str(empty) == ""
    assert not empty.is_valid
    assert empty == none
    assert empty.major is None

    ep = Version("123!2.1+foo.dirty-bar")
    assert str(ep) == "123!2.1+foo.dirty-bar"
    assert ep.pep_440 == "123!2.1+foo.dirty-bar"
    assert ep.is_valid
    assert ep.is_dirty
    assert ep.epoch == 123
    assert ep.main == "2.1"
    assert ep.local_part == "foo.dirty-bar"
    assert ep.mm == "2.1"

    foo = Version("foo")
    assert str(foo) == "foo"
    assert not foo.is_valid
    assert not foo.components
    assert not foo.prerelease
    assert foo.major is None

    bogus = Version("1.2.3.4.5.6")
    assert str(bogus) == "1.2.3.4.5.6"
    assert not bogus.is_valid
    assert not bogus.components
    assert not bogus.prerelease
    assert bogus.mm is None

    v1 = Version("1")
    assert v1.components == (1, 0, 0, 0, 0, 0, "")
    assert str(v1) == "1"
    assert v1.main == "1"
    assert v1.major == 1
    assert v1.minor is None
    assert v1.patch is None
    assert v1.mm is None
    assert empty < v1
    assert v1 > empty
    assert v1 != empty
    assert v1 < 2
    assert v1 < 1.1
    assert v1 < "1.1"
    assert v1 < [5]
    assert v1 > 0
    assert v1 > 0.5
    assert v1 > "0"

    # All versions are bigger than anything not parsing to a valid version
    assert v1 > ""
    assert v1 > []
    assert v1 > [5, 2, 3, 4, 5, 6]

    v1foo = Version("1foo", canonical=None)  # Ignore additional text
    assert v1 == v1foo

    vrc = Version("1.0rc4-bar", canonical=None)
    vdev = Version("1.0a4.dev5-foo", canonical=None)
    assert vdev.pep_440 == "1.0a4.dev5"
    assert vdev.ignored == "-foo"
    assert vrc.is_valid
    assert not vrc.is_final
    assert not vdev.is_final
    assert vdev.prerelease == ("a", 4, "", 0, "dev", 5)
    assert vrc.suffix == "rc"
    assert vdev.suffix == "a.dev"  # Try and convey the fact that we have .a.dev version

    assert vdev < vrc
    assert str(vrc) == "1.0rc4"
    assert str(vdev) == "1.0a4.dev5"
    assert vdev.suffix == "a.dev"
    assert vdev.prerelease == ("a", 4, "", 0, "dev", 5)
    assert vrc.major == 1
    assert vrc.minor == 0
    assert vrc.patch is None
    assert vrc.main == "1.0"

    incomplete_dev = Version("0.4.34dev")
    assert not incomplete_dev.is_final
    assert incomplete_dev.is_valid
    assert incomplete_dev.main == "0.4.34"
    assert incomplete_dev.prerelease == ("", 0, "", 0, "dev", 0)
    assert incomplete_dev.suffix == "dev"

    # Version() can be used in sets/dicts
    s = set()
    s.add(vrc)
    assert vrc in s


def test_version_extraction():
    x = Version.extracted_from_text("foo, version 1.0a4.dev5\nbar baz")
    assert str(x) == "1.0a4.dev5"

    p38 = Version("Python 3.8.6")
    assert not p38.is_valid

    p38 = Version.extracted_from_text("Python 3.8.6")
    assert str(p38) == "3.8.6"
    assert p38.is_valid


def test_version_comparison():
    v10rc5 = Version("1.0rc5")
    assert v10rc5 > "0.9"
    assert v10rc5 > "1.0rc2"
    assert v10rc5 > "1.0rc2.dev10"
    assert v10rc5 > "1.0a5"
    assert v10rc5 < "1.0rc15"
    assert v10rc5 < "1.0"
    assert v10rc5 < "1.1"
    assert v10rc5 < "1.1rc2"

    v = "2.11.1.dev2+b 2.11.0 2.11.1.dev11+a.dirty 1!1.0 2.11.1.dev1 2.11.1"
    v = runez.flattened(v, split=" ", transform=Version)
    assert all(x.is_valid for x in v)
    v = sorted(v)
    v = runez.joined(v)
    assert v == "2.11.0 2.11.1.dev1 2.11.1.dev2+b 2.11.1.dev11+a.dirty 2.11.1 1!1.0"

    v = "1.2rc1 1.2rc2.dev05 1.2rc2.dev4 1.2a1.dev1"
    v = runez.flattened(v, split=" ", transform=Version)
    assert all(x.is_valid for x in v)
    v = sorted(v)
    v = runez.joined(v)
    assert v == "1.2a1.dev1 1.2rc1 1.2rc2.dev4 1.2rc2.dev05"

    v11 = Version("1.1.2.3")
    v12 = Version("1.2.3")
    v12p = Version("1.2.3.post4")
    v2 = Version("2")
    v20 = Version("2.0")
    v20d = Version("2.0.dev1")
    v21d = Version("2.1.dev1")
    v3 = Version("3.0.1.2")
    assert v11.is_final
    assert v12p.is_final
    assert v11.suffix is None
    assert v12p.suffix == "post"
    assert v20d.suffix == "dev"

    # Verify that numerical comparison takes place (not alphanumeric)
    assert None < v12  # For total ordering
    assert v12 > None
    assert v12 == "1.2.3"
    assert v12 == [1, 2, 3]
    assert v12 == (1, 2, 3)
    assert v12 <= (1, 2, 3)
    assert v12 < 1.19
    assert v12 != 1.19
    assert v12 != 1
    assert v12 != (1, 2)
    assert v12 < "1.19"
    assert v12 < "1.19"
    assert v12 != "1.19"
    assert v12 <= "1.19"
    assert v12 <= "1.2.3"
    assert v12 > "1.2"
    assert v12 >= "1.2"
    assert v12 >= "1.2.3"
    assert v12 > 1
    assert v12 == "1.2.3"
    assert v12 < 2
    assert v12 <= 2
    assert v20 == 2
    assert v20 == 2.0
    assert v20 != 2.1

    assert v2 == v20
    assert v12 > v11
    assert v12p > v11
    assert v20 > v11
    assert v20d > v11
    assert v3 > v11
    assert v12p > v12
    assert v20 > v12
    assert v20d > v12
    assert v3 > v12
    assert v20 > v12p
    assert v20d > v12p
    assert v3 > v12p
    assert v21d > v20
    assert v20d < v20
    assert v3 > v20
    assert v3 > v20d


def test_version_crazy():
    verify_ordering([
        Version("1.2.3.7"),
        Version("1.2.3.40dev7"),
        Version("1.2.3.40a6"),
        Version("1.2.3.40rc5.dev0"),
        Version("1.2.3.40c5.dev7"),  # PEP-440 says 'c' should be sorted as 'rc'
        Version("1.2.3.40rc5"),
        Version("1.2.3.40rc5.post6.dev7"),
        Version("1.2.3.40rc5.post6.dev8"),
        Version("1.2.3.40rc5.post6"),
        Version("1.2.3.40rc6.dev7"),
        Version("1.2.3.40rc6"),
        Version("1.2.3.40rc6.post15"),
        Version("1.2.3.40rc20.post6.dev7"),
        Version("1.2.3.40rc20.post17.dev8"),
        Version("1.2.3.40"),
        Version("1.2.3.40post6.dev7"),
        Version("1.2.3.40post6"),
    ])


def test_version_local_part():
    verify_ordering([
        Version("1.0+foo.a"),
        Version("1.0+foo.a.b"),
        Version("1.0+foo.a.5"),
        Version("1.0+foo.z"),
        Version("1.0+foo.z.8"),
        Version("1.0+foo.5"),
        Version("1.0+foo.7"),
    ])


def test_version_ordering():
    verify_ordering([
        Version("1.dev0"),
        Version("1.0.dev456"),
        Version("1.0a1"),
        Version("1.0a2.dev456"),
        Version("1.0a12.dev456"),
        Version("1.0a12"),
        Version("1.0b1.dev456"),
        Version("1.0b2"),
        Version("1.0b2.post345.dev456"),
        Version("1.0b2.post345"),
        Version("1.0rc1.dev456"),
        Version("1.0rc1"),
        Version("1.0"),
        Version("1.0+abc.5"),
        Version("1.0+abc.7"),
        Version("1.0+5"),
        Version("1.0.post456.dev34"),
        Version("1.0.post456"),
        Version("1.0.15"),
        Version("1.1.dev1"),
    ])


def verify_ordering(expected):
    # Jumble the given list of versions a bit, then sort them and verify they sort back to 'expected'
    given = sorted(expected, key=lambda x: x.text)
    mid_item = given.pop(len(given) // 2)
    given.append(mid_item)

    assert given != expected
    assert sorted(given) == expected


def test_version_pep_440():
    vpost = Version("1.2.post")
    assert vpost.is_valid
    assert not vpost.prerelease
    assert str(vpost) == "1.2.post"
    assert vpost.pep_440 == "1.2.post0"

    vrev5 = Version("1.2rev05")
    assert str(vrev5) == "1.2rev05"
    assert vrev5.is_valid
    assert vrev5.pep_440 == "1.2.post5"
    assert vrev5 > vpost

    vrev5_canonical = Version("1.2rev05", canonical=True)
    assert str(vrev5_canonical) == "1.2.post5"
    assert vrev5_canonical.pep_440 == vrev5_canonical.text
    assert vrev5_canonical == vrev5

    vr6 = Version("1.2r6")
    assert str(vr6) == "1.2r6"
    assert vr6.is_valid
    assert vr6.pep_440 == "1.2.post6"
    assert vr6 > vrev5

    vrev5dev3 = Version("1.2rev05.dev3")
    assert vrev5dev3.prerelease
    assert str(vrev5dev3) == "1.2rev05.dev3"
    assert vrev5dev3.pep_440 == "1.2.rev5.dev3.post5"
    assert vpost < vrev5dev3
    assert vrev5 > vrev5dev3
    assert vr6 > vrev5dev3

    vrc1 = Version("v1.39.4-rc.1")
    assert str(vrc1) == "1.39.4-rc.1"
    assert vrc1.pep_440 == "1.39.4rc1"



from runez.serialize import read_json, save_json
from runez.system import _R, resolved_path, stringified, SYS_INFO


def ask_once(name, instructions, default=None, base="~/.config", serializer=stringified, fatal=False, logger=False):
    """
    Args:
        name (str): Name under which to store provided answer (will be stored in ~/.config/<name>.json)
        instructions (str): Instructions to show to user when prompt is necessary
        default: Default value to return if answer not available
        base (str | pathlib.Path): Base folder where to stored provided answer
        serializer (callable): Function that will turn provided value into object to be stored
        logger (callable | bool | None): Logger to use, True to print(), False to trace(), None to disable log chatter
        fatal (type | bool | None): True: abort execution on failure, False: don't abort but log, None: don't abort, don't log

    Returns:
        Value given by user (or 'default' if given), optionally wrapped via `serializer`
    """
    path = resolved_path(name, base=base)
    if not path.endswith(".json"):
        path += ".json"

    existing = read_json(path, logger=logger)
    if existing is not None:
        return existing

    if not SYS_INFO.terminal.is_stdout_tty:
        return _R.habort(default, fatal, logger, "Can't prompt for %s, not on a tty" % name)

    try:
        provided = input(instructions)
        if provided:
            value = serializer(provided)
            if value is not None:
                save_json(value, path, fatal=fatal, logger=logger)
                return value

            return _R.habort(default, fatal, logger, "Invalid value provided for %s" % name)

        return _R.habort(default, fatal, logger, "No value provided for %s" % name)

    except KeyboardInterrupt:
        return _R.habort(default, fatal, logger, "Cancelled by user")



import errno
import logging
import os
import subprocess
import sys
from unittest.mock import patch

import pytest

import runez
from runez.conftest import exception_raiser
from runez.program import RunAudit, RunResult


CHATTER = runez.DEV.tests_path("chatter")


def simulate_os_error(code):
    e = OSError(code)
    e.errno = code

    def do_raise(*_):
        raise e

    return do_raise


def test_background_run(logged):
    with runez.CurrentFolder(os.path.dirname(CHATTER)):
        r = runez.run(CHATTER, "hello", background=True, dryrun=True, logger=True)
        assert r.succeeded
        assert "chatter hello &" in logged.pop()

        r = runez.run(CHATTER, "hello", background=True, dryrun=False)
        assert r.succeeded
        assert r.pid
        assert r.output is None
        assert r.error is None
        assert "chatter hello &" in logged.pop()


class CrashingWrite:
    """Simulate a file/stream that keeps on crashing"""

    crash_counter = 0

    def write(self, message):
        self.crash_counter += 1
        raise Exception("oops, failed to write %s" % message)


@pytest.mark.skipif(runez.SYS_INFO.platform_id.is_windows, reason="Not supported on windows")
def test_capture(monkeypatch):
    with runez.CurrentFolder(os.path.dirname(CHATTER)):
        # Check which finds programs in current folder
        assert runez.which("chatter") == CHATTER
        assert runez.shell("chatter hello") == "hello"

    with runez.CaptureOutput(dryrun=True) as logged:
        # Dryrun mode doesn't fail (since it doesn't actually run the program)
        r = runez.run(CHATTER, "silent-fail", fatal=True)
        assert r.succeeded
        assert "[dryrun] " in r.output
        assert r.error == ""
        assert "Would run:" in logged.pop()

        r = runez.run(CHATTER, "silent-fail", stdout=None, stderr=None, fatal=True)
        assert r.succeeded
        assert r.output is None
        assert r.error is None
        assert "Would run:" in logged.pop()

    with runez.CaptureOutput(seed_logging=True) as logged:
        # Test success
        assert runez.run(CHATTER, "hello", fatal=False) == RunResult("hello", "", 0)
        assert runez.run(CHATTER, "hello", fatal=True) == RunResult("hello", "", 0)
        assert "chatter hello" in logged.pop()
        assert runez.run(CHATTER, stdout=None) == RunResult(None, "", 0)
        assert "Running:" in logged.pop()
        crasher = CrashingWrite()
        assert runez.run(CHATTER, "hello", fatal=True, passthrough=crasher) == RunResult("hello", "", 0)
        assert crasher.crash_counter
        assert "hello" in logged.pop()

        # Test no-wait
        r = runez.run(CHATTER, "hello", fatal=None, stdout=None, stderr=None)
        assert r.exit_code is None  # We don't know exit code because we didn't wait
        assert r.pid

        r = runez.run(CHATTER, stdout=None, stderr=None)
        assert r
        assert str(r) == "RunResult(exit_code=0)"
        assert r.succeeded
        assert r.output is None
        assert r.error is None
        assert r.full_output is None

        r = runez.run(CHATTER, "hello", path_env={"PATH": ":.", "CPPFLAGS": " -I/usr/local/opt/openssl/include"})
        assert str(r) == "RunResult(exit_code=0)"
        assert r.succeeded
        assert r.output == "hello"
        assert r.error == ""
        assert r.full_output == "hello"

        # Test stderr
        r = runez.run(CHATTER, "complain")
        assert r.succeeded
        assert r.output == ""
        assert r.error == "complaining"
        assert r.full_output == "complaining"
        logged.pop()

        # Test failure
        with pytest.raises(Exception):
            runez.run(CHATTER, "fail")
        assert "Run failed:" in logged.pop()

        r = runez.run(CHATTER, "silent-fail", fatal=False)
        assert str(r) == "RunResult(exit_code=1)"
        assert r.failed
        assert r.error == ""
        assert r.output == ""
        assert r.full_output == r.error

        if hasattr(subprocess.Popen, "__enter__"):
            # Simulate an EIO
            with patch("runez.program._read_data", side_effect=simulate_os_error(errno.EIO)):
                r = runez.run(CHATTER, "fail", fatal=False, passthrough=True)
                assert r.failed
                assert r.exc_info is None
                assert r.output == ""
                assert r.error == ""

            # Simulate an OSError
            with patch("runez.program._read_data", side_effect=simulate_os_error(errno.EINTR)):
                r = runez.run(CHATTER, "fail", fatal=False, passthrough=True)
                assert r.failed
                assert r.output is None
                assert "failed: OSError(" in r.error

        # Verify no extra "exited with code ..." message is added when pass-through had some output
        logged.clear()
        with pytest.raises(SystemExit):
            runez.run(CHATTER, "fail", fatal=SystemExit, passthrough=True)
        assert "exited with code" not in logged.pop()

        with pytest.raises(runez.system.AbortException):
            runez.run(CHATTER, "fail", fatal=True, passthrough=True)
        assert "exited with code" not in logged.pop()

        # Verify that silent pass-through gets at least mention of exit code
        with pytest.raises(SystemExit):
            runez.run(CHATTER, "silent-fail", fatal=SystemExit, passthrough=True)
        assert "exited with code" in logged.pop()

        with pytest.raises(runez.system.AbortException):
            runez.run(CHATTER, "silent-fail", fatal=True, passthrough=True)
        assert "exited with code" in logged.pop()

        r = runez.run(CHATTER, "fail", fatal=False, passthrough=True)
        assert r.failed
        assert r.error == "failed"
        assert r.output == "hello there"
        assert r.full_output == "failed\nhello there"

        r = runez.run("foo/bar", fatal=False)
        assert r.exit_code == 1
        assert "foo/bar is not an executable" in r.error

        r = runez.run("foo-bar-no-such-program", fatal=False)
        assert r.exit_code == 1
        assert "is not installed (PATH=" in r.error

        with monkeypatch.context() as m:
            m.setattr(subprocess, "Popen", exception_raiser(OSError("testing")))
            r = runez.run("python", "--version", fatal=False)
            assert not r
            assert r.failed
            assert "python failed: OSError(" in r.error
            assert r.output is None

            with pytest.raises(OSError):
                runez.run("python", "--version")

        # Test convenience arg None filtering
        logged.clear()
        assert runez.run(CHATTER, "hello", "-a", 0, "-b", None, 1, 2, None, "foo bar") == RunResult("hello -a 0 1 2 foo bar", "", 0)
        assert 'chatter hello -a 0 1 2 "foo bar"' in logged.pop()


@patch("runez.program.os.fork", return_value=None)
@patch("runez.program.os.setsid")
@patch("runez.program.os.open")
@patch("runez.program.os.dup2")
@patch("runez.program.os.close")
def test_daemonize(*_):
    # This simply exercises code daemonize() that would otherwise run in a forked process
    assert runez.program.daemonize() is None


@pytest.mark.skipif(runez.SYS_INFO.platform_id.is_windows, reason="Not supported on windows")
def test_executable(temp_folder):
    with runez.CaptureOutput(dryrun=True) as logged:
        assert runez.make_executable("some-file") == 1
        assert "Would make some-file executable" in logged.pop()
        assert runez.make_executable("some-file", logger=False) == 1
        assert not logged

    with runez.CaptureOutput() as logged:
        assert runez.touch("some-file") == 1
        assert "Touched some-file" in logged.pop()
        assert runez.delete("some-file") == 1
        assert "Deleted some-file" in logged.pop()
        assert runez.touch("some-file", logger=logging.debug) == 1
        assert "Touched some-file" in logged.pop()
        assert runez.make_executable("some-file", logger=logging.debug) == 1
        assert "Made 'some-file' executable" in logged.pop()
        assert runez.is_executable("some-file")
        assert runez.make_executable("some-file") == 0
        assert not logged

        assert runez.touch("some-file", logger=False) == 1
        assert runez.delete("some-file", logger=False) == 1
        assert not runez.is_executable("some-file")
        assert not logged

        assert runez.make_executable("/dev/null/some-file", fatal=False) == -1
        assert "does not exist, can't make it executable" in logged.pop()

        assert runez.make_executable("/dev/null/some-file", fatal=False, logger=None) == -1  # Don't log anything
        assert not logged

        assert runez.make_executable("/dev/null/some-file", fatal=False, logger=False) == -1  # Log errors only
        assert "does not exist, can't make it executable" in logged.pop()


def test_pids():
    assert not runez.check_pid(None)
    assert not runez.check_pid(0)
    assert not runez.check_pid("foo")  # noqa, garbage given, don't crash

    assert runez.check_pid(os.getpid())
    assert not runez.check_pid(1)  # No privilege to do this (tests shouldn't run as root)


def check_process_tree(pinfo, max_depth=10):
    """Verify that process info .parent does not recurse infinitely"""
    if pinfo:
        assert max_depth > 0
        check_process_tree(pinfo.parent, max_depth=max_depth - 1)


def test_ps():
    assert runez.PsInfo.from_pid(None) is None
    assert runez.PsInfo.from_pid(0) is None

    p = runez.PsInfo()
    check_process_tree(p)
    assert p == runez.PsInfo(0)
    assert p == runez.PsInfo("0")
    assert p == runez.PsInfo(os.getpid())
    assert p == runez.PsInfo("%s" % os.getpid())

    info = p.info
    assert info["PID"] in str(p)
    assert p.cmd
    assert p.cmd_basename
    assert p.ppid == os.getppid()
    assert p.userid != p.uid

    parent = p.parent
    assert parent.pid == p.ppid

    # Verify that both variants (user name or uid number) for UID work
    uid = p.uid
    userid = p.userid
    p = runez.PsInfo()
    if runez.to_int(info["UID"]) is None:
        p.info["UID"] = uid

    else:
        p.info["UID"] = userid

    assert p.uid == uid
    assert p.userid == userid

    # Edge case: verify __eq__ based on pid
    p.pid = 0
    assert p != runez.PsInfo(0)


def simulated_ps_output(pid, ppid, cmd):
    template = "UID   PID  PPID CMD\n  0 {pid:>5} {ppid:>5} {cmd}"
    return RunResult(output=template.format(pid=pid, ppid=ppid, cmd=cmd), code=0)


def simulated_tmux(program, *args, **_):
    if program == "tmux":
        return RunResult(output="3", code=0)

    if program == "id":
        if args[0] == "-un":
            return RunResult(output="root", code=0)

        return RunResult(output="0", code=0)

    assert program == "ps"
    pid = args[1]
    if pid == 1:
        return simulated_ps_output(pid, 0, "/sbin/init")

    if pid == 2:
        return simulated_ps_output(pid, 1, "tmux new-session ...")

    if pid == 3:
        return simulated_ps_output(pid, 1, "tmux attach-session ...")

    if pid == -1:
        return RunResult(code=1)

    return simulated_ps_output(pid, 2, "/dev/null/some-test foo bar")


def test_ps_follow():
    with patch("runez.program.run", side_effect=simulated_tmux):
        assert runez.PsInfo.from_pid(-1) is None
        bad_pid = runez.PsInfo(-1)
        assert str(bad_pid) == "-1 None None"
        assert bad_pid.cmd is None
        assert bad_pid.cmd_basename is None
        assert bad_pid.info is None
        assert bad_pid.followed_parent is None
        assert bad_pid.parent is None
        assert bad_pid.pid == -1
        assert bad_pid.ppid is None
        assert bad_pid.uid is None
        assert bad_pid.userid is None
        assert bad_pid.parent_list(follow=True) == []
        assert bad_pid.parent_list(follow=False) == []

        p = runez.PsInfo()
        assert p.cmd == "/dev/null/some-test foo bar"
        assert p.cmd_basename == "/dev/null/some-test"  # Falls back to using 1st sequence with space as basename
        assert p.uid == 0
        assert p.userid == "root"
        parents = p.parent_list(follow=False)
        followed_parents = p.parent_list(follow=True)

        # Verify that parent_list(follow=True) follows parent processes properly
        assert parents != followed_parents
        assert parents == [p.parent, p.parent.parent]
        assert followed_parents == [p.followed_parent, p.followed_parent.parent]

        with patch("runez.program.is_executable", side_effect=lambda x: x == "/dev/null/some-test foo"):
            # Edge case: verify that `ps` lack of quoting is properly detected
            p = runez.PsInfo()
            assert p.cmd == "/dev/null/some-test foo bar"
            assert p.cmd_basename == "some-test foo"


def check_ri(platform, instructions=None):
    with pytest.raises(Exception) as exc:
        runez.program.require_installed("foo", instructions=instructions, platform=platform)

    return str(exc)


def test_require_installed(monkeypatch):
    monkeypatch.setattr(runez.program, "which", lambda x: "/bin/foo")
    assert runez.program.require_installed("foo") is None  # Does not raise

    monkeypatch.setattr(runez.program, "which", lambda x: None)
    r = check_ri("macos")
    assert "foo is not installed, run: `brew install foo`" in r

    r = check_ri("linux")
    assert "foo is not installed, run: `apt install foo`" in r

    r = check_ri("macos", instructions="custom instructions")
    assert "foo is not installed, custom instructions" in r

    r = check_ri("unknown-platform")
    assert "foo is not installed:" in r
    assert "- on macos: run: `brew install foo`" in r
    assert "- on linux: run: `apt install foo`" in r


def test_run_description():
    short_py = runez.short(sys.executable)
    audit = RunAudit(sys.executable, ["-mpip", "--help"], {})
    assert str(audit) == "pip --help"
    assert audit.run_description() == "pip --help"
    assert audit.run_description(short_exe=None) == "%s -mpip --help" % short_py
    assert audit.run_description(short_exe=False) == "%s -mpip --help" % short_py
    assert audit.run_description(short_exe=True) == "pip --help"
    assert audit.run_description(short_exe="foo") == "foo -mpip --help"

    audit = RunAudit(sys.executable, ["-m", "pip", "--help"], {})
    assert audit.run_description() == "pip --help"
    assert audit.run_description(short_exe=None) == "%s -m pip --help" % short_py
    assert audit.run_description(short_exe=False) == "%s -m pip --help" % short_py
    assert audit.run_description(short_exe=True) == "pip --help"
    assert audit.run_description(short_exe="foo") == "foo -m pip --help"

    audit = RunAudit(sys.executable, ["bin/pip/__main__.py", "--help"], {})
    assert audit.run_description() == "pip --help"
    assert audit.run_description(short_exe=None) == "%s bin/pip/__main__.py --help" % short_py
    assert audit.run_description(short_exe=False) == "%s bin/pip/__main__.py --help" % short_py
    assert audit.run_description(short_exe=True) == "pip --help"
    assert audit.run_description(short_exe="foo") == "foo bin/pip/__main__.py --help"

    audit = RunAudit("foo/python3", ["-mpip", "--help"], {})
    assert audit.run_description() == "foo/python3 -mpip --help"
    assert audit.run_description(short_exe=None) == "foo/python3 -mpip --help"
    assert audit.run_description(short_exe=False) == "foo/python3 -mpip --help"
    assert audit.run_description(short_exe=True) == "pip --help"
    assert audit.run_description(short_exe="foo") == "foo -mpip --help"

    audit = RunAudit("foo/bar", ["-mpip", "--help"], {})
    assert audit.run_description() == "foo/bar -mpip --help"
    assert audit.run_description(short_exe=None) == "foo/bar -mpip --help"
    assert audit.run_description(short_exe=False) == "foo/bar -mpip --help"
    assert audit.run_description(short_exe=True) == "foo/bar -mpip --help"
    assert audit.run_description(short_exe="foo") == "foo -mpip --help"

    cmd = runez.to_path(runez.SYS_INFO.venv_bin_path("foo"))
    audit = RunAudit(cmd, ["--help"], {})
    assert str(audit) == "foo --help"
    assert audit.run_description() == "foo --help"


def test_which():
    assert runez.which(None) is None
    assert runez.which("/dev/null") is None
    assert runez.which("dev/null") is None
    pp = runez.which(runez.to_path("python"))
    ps = runez.which("python")
    assert pp == ps


@pytest.mark.skipif(runez.SYS_INFO.platform_id.is_windows, reason="Not supported on windows")
def test_wrapped_run(monkeypatch):
    original = ["python", "-mvenv", "foo"]
    monkeypatch.delenv("PYCHARM_HOSTED", raising=False)
    with runez.program._WrappedArgs(original) as args:
        assert args == original

    monkeypatch.setenv("PYCHARM_HOSTED", "1")
    with runez.program._WrappedArgs(original) as args:
        assert args
        assert len(args) == 5
        assert args[0] == "/bin/sh"
        assert os.path.basename(args[1]) == "pydev-wrapper.sh"



import os
from unittest.mock import patch

import pytest

import runez
from runez.prompt import ask_once


def custom_serializer(value):
    if value == "invalid":
        return None

    return {"value": value}


def mocked_input(x):
    return x


def test_no_tty():
    assert ask_once("test", "Please enter value: ") is None
    with pytest.raises(runez.system.AbortException):
        ask_once("test", "Please enter value: ", fatal=True)


def test_with_tty(monkeypatch, logged):
    with patch("runez.prompt.input", side_effect=mocked_input):
        monkeypatch.setattr(runez.SYS_INFO.terminal, "is_stdout_tty", True)
        expected = {"value": "foo"}
        with runez.TempFolder() as tmp:
            assert ask_once("test", "foo", base=tmp, serializer=custom_serializer) == expected

            assert runez.read_json("test.json", logger=None) == expected
            assert ask_once("test", "bar", base=tmp) == expected  # Ask a 2nd time, same response

            # Verify that if `serializer` returns None, value is not returned/stored
            with pytest.raises(Exception) as exc:
                ask_once("test-invalid", "invalid", base=tmp, serializer=custom_serializer, fatal=True, logger=None)
            assert "Invalid value provided for test-invalid" in str(exc)
            assert not os.path.exists("test-invalid.json")

            # Same, but don't raise exception (returns default)
            assert ask_once("test-invalid", "invalid", base=tmp, serializer=custom_serializer) is None
            assert not logged  # Traced by default

            # Simulate no value provided
            with pytest.raises(Exception) as exc:
                ask_once("test-invalid", "", base=tmp, serializer=custom_serializer, fatal=True)
            assert "No value provided" in str(exc)
            assert "No value provided" in logged.pop()  # Logged if fatal=True

    with patch("runez.prompt.input", side_effect=KeyboardInterrupt):
        # Simulate CTRL+C
        with pytest.raises(Exception) as exc:
            ask_once("test2", "test2", base=tmp, serializer=custom_serializer, fatal=True)
        assert "Cancelled by user" in str(exc)



"""
Test click related methods
"""

import errno
import logging
import os
import sys

import click
import pytest

import runez
from runez.conftest import exception_raiser


def my_formatter(text):
    return text.format(placeholder="epilog")


@runez.click.group()
@runez.click.version(message="%(prog)s, version %(version)s")
@runez.click.color()
@runez.click.config("-c", prefix="g.")
@runez.click.debug()
@runez.click.dryrun()
@runez.click.log(expose_value=False)
def my_group(debug):
    # By default, --config value is NOT exposed, the global runez.config.CONFIG is altered
    runez.system.AbortException = SystemExit
    runez.log.setup(
        debug=debug,
        console_format="%(levelname)s %(message)s",
        console_level=logging.INFO,
        locations=None,
        greetings=":: {argv}",
    )
    config = runez.config.CONFIG
    cd = config.get("g.cd")
    if cd:
        logging.info("Changed folder to %s" % runez.short(cd))
        runez.ensure_folder(cd)
        os.chdir(cd)

    assert len(config.providers) == 1
    assert "--config: " in config.overview()
    assert config.provider_by_name("--config") is not None


@my_group.command()
@click.argument("text", nargs=-1)
def echo(text):
    """
    Repeat provided text

    This part will be an {placeholder}
    """
    text = " ".join(text)
    if text == "AssertionError":
        assert False, "oops"

    msg = "%s, color: %s, %s values, g.a=%s" % (text, runez.color.is_coloring(), len(runez.config.CONFIG), runez.config.get("g.a"))
    msg += ", debug: %s, dryrun: %s, log: %s" % (runez.log.debug, runez.DRYRUN, runez.log.spec.file_location)
    print(msg)


@runez.click.command()
@runez.click.version(version="1.2.3")
@runez.click.border("-b")
@runez.click.color("-x", expose_value=True)
@runez.click.config("-c", expose_value=True, default="a=b,c=d", split=",", env="MY_PROG", propsfs=True)
@runez.click.debug()
@runez.click.dryrun()
@runez.click.log()
@click.option("--use-stderr", is_flag=True, help="Print on stderr")
def say_hello(border, color, config, debug, log, use_stderr):
    """Say hello"""
    # When --config is exposed, global config is NOT modified
    assert not runez.config.CONFIG.providers
    assert runez.log.spec.file_location is None
    # Intentionally set global runez.config.CONFIG to verify it has been restored at the end of the test command run
    runez.config.CONFIG = config
    assert len(config.providers) == 4
    assert "propsfs" in config.overview()
    msg = "border: %s, color: %s, a=%s c=%s, debug: %s, dryrun: %s, log: %s" % (
        border, color, config.get("a"), config.get("c"), debug, runez.DRYRUN, log
    )
    if use_stderr:
        sys.stderr.write("%s\n" % msg)

    else:
        print(msg)


def test_command(cli, monkeypatch):
    cli.main = say_hello
    cli.run("--help")
    assert cli.succeeded
    assert "--border" in cli.logged
    assert "--color" in cli.logged
    assert "--no-color" in cli.logged

    cli.expect_success("--version", "1.2.3")
    cli.expect_success(["--help"], "-x, --color / --no-color", "--log PATH", "Say hello")

    cli.run("--no-color")
    assert cli.succeeded
    cli.assert_printed("border: reddit, color: False, a=b c=d, debug: None, dryrun: False, log: None")

    cli.run("-x")
    assert cli.succeeded
    assert "color: True" in cli.logged.stdout

    cli.run("--border github --color --debug --config=a=x -c c=y --log=foo")
    assert cli.succeeded
    cli.assert_printed("border: github, color: True, a=x c=y, debug: True, dryrun: False, log: foo")

    monkeypatch.setenv("MY_PROG_A", "some-value")
    monkeypatch.delenv("LANG", raising=False)
    cli.run("")
    assert cli.succeeded
    cli.assert_printed("border: reddit, color: None, a=some-value c=d, debug: None, dryrun: False, log: None")


def test_command_with_stderr(cli):
    cli.main = say_hello
    cli.run("")
    assert cli.succeeded
    assert "border:" in cli.logged

    cli.main = say_hello
    cli.run("--use-stderr")
    assert cli.succeeded
    assert "border:" in cli.logged


def sample_config(**attrs):
    attrs.setdefault("tracer", print)
    c = runez.click._ConfigOption(attrs)
    return c


def test_config(isolated_log_setup, logged, monkeypatch):
    # sys.argv is used as env var prefix when env=True is used
    runez.log.enable_trace(True)
    config = sample_config(env=True)(None, None, "")
    assert str(config) == "--config, PYTEST_* env vars"
    assert "Adding config provider PYTEST_*" in logged.pop()

    monkeypatch.setenv("MY_PROG_A", "via env")
    propsfs = runez.DEV.tests_path("sample")
    config = sample_config(env="MY_PROG", default="x=y", propsfs=propsfs, split=",")
    c1 = config(None, None, "")
    assert str(c1) == "--config, MY_PROG_* env vars, propsfs, --config default"
    assert logged.pop()
    assert c1.get("x") == "y"
    assert "Using x='y' from --config default" in logged.stderr.pop()
    assert c1.get_int("some-int") == 123
    assert "Using some-int='123' from propsfs" in logged.stderr.pop()

    # 'some-int' from propsfs sample is overridden
    c2 = config(None, None, "x=overridden,some-int=12,twenty-k=20kb,five-one-g=5.1g")
    assert logged.pop()
    assert c1.get_str("key") is None
    assert not logged
    assert c2.get("x") == "overridden"
    assert "Using x='overridden' from --config" in logged.stderr.pop()
    assert c2.get_int("some-int") == 12
    assert "Using some-int='12' from --config" in logged.stderr.pop()

    # Test bytesize
    assert c2.get_bytesize("some-int") == 12
    assert c2.get_bytesize("some-int", default_unit="k") == 12 * 1024
    assert c2.get_bytesize("some-int", default_unit="m") == 12 * 1024 * 1024

    assert c2.get_bytesize("twenty-k") == 20 * 1024
    assert c2.get_bytesize("five-one-g") == int(5.1 * 1024 * 1024 * 1024)

    assert c2.get_bytesize("twenty-k", minimum=5, maximum=100) == 100

    # Invalid default unit affects only ints without unit
    assert c2.get_bytesize("some-int", default_unit="a") is None
    assert c2.get_bytesize("twenty-k", default_unit="a") == 20 * 1024

    assert c2.get_bytesize("some-string") is None
    assert c2.get_bytesize("some-string", default=5) == 5
    assert c2.get_bytesize("some-string", default="5k") == 5 * 1024
    assert c2.get_bytesize("some-string", default=5, default_unit="k") == 5 * 1024
    assert c2.get_bytesize("some-string", default="5m", default_unit="k") == 5 * 1024 * 1024
    logged.pop()

    # Shuffle things around
    c2.add(c2.providers[0])
    assert "Replacing config provider --config at index 0" in logged.stderr.pop()

    c2.add(runez.config.DictProvider({}, name="foo1"), front=True)
    assert "Adding config provider foo1 to front" in logged.stderr.pop()


def test_group(cli):
    assert runez.system.AbortException is not SystemExit
    cli.main = my_group
    runez.click.prettify_epilogs(my_group, formatter=my_formatter)
    runez.click.prettify_epilogs(my_group, formatter=my_formatter)  # Calling this multiple times is a no-op
    cli.expect_success("--version", "my-group, version ")

    cli.run("--help")
    assert cli.succeeded
    assert "--color / --no-color" in cli.logged
    assert "--log PATH" in cli.logged
    assert "Repeat provided text" in cli.logged
    assert "This part will be" not in cli.logged

    assert cli.context == os.getcwd()
    cli.run("-ccd=foo", "echo", "--help")
    assert cli.succeeded
    assert "Repeat provided text" in cli.logged
    assert "This part will be an epilog" in cli.logged
    assert "Changed folder to foo" in cli.logged  # TODO: correct it so it doesn't get logged twice
    # Verify that current folder was restored
    assert cli.context == os.getcwd()
    assert runez.system.AbortException is not SystemExit

    cli.run("--color echo hello")
    assert cli.succeeded
    cli.assert_printed("hello, color: True, 0 values, g.a=None, debug: False, dryrun: False, log: None")

    cli.run("--no-color --dryrun -ca=b --config c=d --log foo echo hello")
    assert cli.succeeded
    cli.assert_printed("hello, color: False, 2 values, g.a=b, debug: False, dryrun: True, log: foo")

    with pytest.raises(AssertionError):
        cli.run("echo", "AssertionError")


def check_protected_main(exit_code, exception, *messages, **kwargs):
    with runez.CaptureOutput() as logged:
        with pytest.raises(SystemExit) as x:
            runez.click.protected_main(exception_raiser(exception), **kwargs)
        assert x.value.code == exit_code
        for message in messages:
            assert message in logged

        return logged


def test_protected_main():
    # Exercise protected_main()
    check_protected_main(1, KeyboardInterrupt, "Aborted")
    check_protected_main(1, NotImplementedError, "Not implemented yet")
    check_protected_main(1, Exception("oops"), "Exception: oops", "Traceback")

    # No stack trace with debug_stacktrace=True
    logged = check_protected_main(1, Exception("oops"), "oops", debug_stacktrace=True)
    assert "Traceback" not in logged

    check_protected_main(1, TypeError("oops"), "TypeError",  "Traceback", no_stacktrace=[ValueError])
    logged = check_protected_main(1, ValueError("oops"), "oops", no_stacktrace=[ValueError])
    assert "ValueError" not in logged  # Exception is stringified and shown as ERROR
    assert "Traceback" not in logged

    exc = IOError()
    exc.errno = errno.EPIPE
    logged = check_protected_main(0, exc)
    assert not logged


def test_settings():
    s = runez.click.settings(foo="bar", epilog="some epilog")
    assert len(s) == 3
    assert s["epilog"] == "some epilog"
    assert s["foo"] == "bar"
    assert s["context_settings"] == dict(help_option_names=["-h", "--help"], max_content_width=140)

    s = runez.click.settings(help="-h --help --explain")
    assert s["context_settings"]["help_option_names"] == ["-h", "--help", "--explain"]



"""
Simple REST client with --dryrun support, and optionally 'fatal=...', 'logger=...' (same as all other IO related runez functions).

Benefits and use case:
- Convenient for short-run CLIs, who typically don't do anything sophisticated (with threads etc) when accessing REST services
- Simplified code (no need to repeat session = requests.Session() ... etc)
- Convenience touches like 'base_url'
- Can support different implementations (anything that has a requests-like API will do: session.get(), .post() etc)
- Simple mocks for tests

Usage pattern:
    from runez.http import RestClient

    # You can replace the default RestClient.handler
    RestClient.handler = MyHandler
    MY_CLIENT = runez.http.RestClient("https://example.net", foo="my-foo")

    # Or do it per client:
    MY_CLIENT = runez.http.RestClient("https://example.net", spec="my-spec", handler=MyHandler)

    # Then use the client
    response = MY_CLIENT.get("api/v1/....", fatal=False, dryrun=False)
"""

import functools
import json
import os
import re
import sys
import urllib.parse
from pathlib import Path

from runez.file import checksum, decompress, delete, ensure_folder, TempFolder, to_path
from runez.system import _R, abort, DEV, find_caller, joined, short, stringified, SYS_INFO, UNSET


def urljoin(base, url):
    """Join base + url, considering that `base` is intended to be the base url of a REST end point"""
    if not base:
        return url

    if not url:
        return base

    if not base.endswith("/"):
        # Annoying edge case: urljoin() removes the last component of url if it's not `/` terminated, which can be surprising
        base += "/"

    return urllib.parse.urljoin(base, url, allow_fragments=False)


class CacheWrapper:
    """
    Simple cache wrapper, intended to provide a convenient way of caching REST calls via pypi module diskcache for example
    By default, ~/.cache/{program_name} is used

    Example usage:
        cache = RestClient.std_diskcache(size_limit="4g")
        client = RestClient(url, cache=cache)
    """

    base_location = "~/.cache/{program_name}"
    cachable_methods = {"GET": True, "POST": True}  # By default, cache only these REST methods
    default_expire = "1h"  # 1 hour
    size_limit = "2g"  # 2 GB

    __dev_suffix = "-dev"  # Override to None to disable distinguishing dev runs from non-dev runs

    @classmethod
    def resolved_base_location(cls, base_location=UNSET):
        if base_location is UNSET:
            base_location = cls.base_location

        if base_location:
            base_location = base_location.format(program_name=SYS_INFO.program_name)
            return os.path.expanduser(base_location)

    @classmethod
    def cache_base_path(cls, base_location=UNSET, suffix=None, dev_suffix=UNSET):
        """
        Args:
            base_location (str | None): Base location to use, {program_name}
            suffix (str | None): Optional suffix for location basename to add
            dev_suffix (str | None): Suffix to add for detected development runs (from a dev venv)

        Returns:
            (str): Fully resolved location, example: ~/.cache/my-program-py39-dev
        """
        path = cls.resolved_base_location(base_location)
        if path:
            if suffix:
                path += "-%s" % suffix

            if dev_suffix is UNSET:
                dev_suffix = cls.__dev_suffix

            if dev_suffix and DEV.venv_folder:
                # Allows to distinguish development runs (dev and "prod" runs won't use the same cache)
                path += dev_suffix

        return path

    def __init__(self, cache_backend, base_location, default_expire, size_limit):
        """
        Args:
            cache_backend: Underlying backend to use
            base_location (str | None): Cache base location
            default_expire (int | float | None): Default expiration time in seconds
            size_limit (int | None): Max size in bytes for this cache
        """
        self.base_location = base_location
        self.cachable_methods = dict(self.cachable_methods)
        self.default_expire = _R.lc.rm.to_seconds(default_expire)
        self.size_limit = _R.lc.rm.to_bytesize(size_limit)
        self.cache_backend = cache_backend

    @staticmethod
    def cache_key(full_url, params=None):
        """
        Args:
            full_url (str): Full URL to remote asset
            params (dict | None): Optional REST query params

        Returns:
            (str): Key to use to represent this remote asset in the cache
        """
        key = full_url
        if params:
            key += "?%s" % urllib.parse.urlencode(params)

        return key

    def is_cachable_method(self, method):
        """
        Args:
            method (str): Should this method be cached (REST methods: DELETE, GET, HEAD, POST, PURGE, PUT)

        Returns:
            (bool): True if this REST method call should be cached
        """
        return self.cachable_methods.get(method)

    def delete(self, cache_key):
        """
        Args:
            cache_key (str): Key to delete from cache
        """
        self.cache_backend.delete(cache_key)

    def get(self, cache_key):
        """
        Args:
            cache_key (str): Key to lookup from this cache
        """
        return self.cache_backend.get(cache_key)

    def set(self, cache_key, data, expire=UNSET):
        """
        Args:
            cache_key (str): Key to store `data` under
            data: Data to store
            expire (int | float | None): Expiration time in seconds
        """
        if expire is UNSET:
            expire = self.default_expire

        return self.cache_backend.set(cache_key, data, expire=expire)


class ForbiddenHttpError(Exception):
    """Raised to signify test setup prevented a remote call"""


class GlobalHttpCalls:
    """Allows to forbid/allow outgoing http(s) call during test runs"""

    _original_urlopen = None

    def __init__(self, allowed):
        """We're used as a context manager"""
        self.allowed = allowed
        self.was_allowed = None

    def __repr__(self):
        return "allowed" if self.allowed else "forbidden"

    def __enter__(self):
        self.was_allowed = self.allow(self.allowed)
        return self

    def __exit__(self, *_):
        self.allow(self.was_allowed)

    @staticmethod
    def intentionally_disabled(*_, **kwargs):
        """Used as replacement of urlopen(), when external http calls are forbidden"""
        raise ForbiddenHttpError(kwargs.get("url"))

    @classmethod
    def is_forbidden(cls):
        """Are outgoing http calls currently allowed?"""
        return cls._original_urlopen is not None

    @classmethod
    def allowed(cls, func):
        """Decorator for test_ functions, temporarily allows external requests"""
        @functools.wraps(func)
        def inner(*args, **kwargs):
            with GlobalHttpCalls(True):
                return func(*args, **kwargs)

        return inner

    @classmethod
    def forbidden(cls, func):
        """Decorator for test_ functions, temporarily forbids external requests"""
        @functools.wraps(func)
        def inner(*args, **kwargs):
            with GlobalHttpCalls(False):
                return func(*args, **kwargs)

        return inner

    @classmethod
    def allow(cls, allowed=True):
        """Allow outgoing http(s) calls"""
        from urllib3.connectionpool import HTTPConnectionPool

        was_allowed = cls._original_urlopen is None
        if allowed:
            if cls._original_urlopen is not None:
                HTTPConnectionPool.urlopen = cls._original_urlopen
                cls._original_urlopen = None

        elif cls._original_urlopen is None:
            cls._original_urlopen = HTTPConnectionPool.urlopen
            HTTPConnectionPool.urlopen = GlobalHttpCalls.intentionally_disabled

        return was_allowed

    @classmethod
    def forbid(cls):
        """Forbid outgoing http(s) calls"""
        return cls.allow(False)


class DataState:
    """Model data/files for POST/PUT requests only"""

    data = None
    json = None
    files = None
    fhandles = None

    def add_fh(self, path, dryrun):
        if _R.resolved_dryrun(dryrun):
            return path

        if self.fhandles is None:
            self.fhandles = []

        fh = open(path, mode="rb")
        self.fhandles.append(fh)
        return fh

    def complete(self, keyword_args):
        for key in ("data", "json", "files"):
            value = getattr(self, key)
            if value:
                keyword_args[key] = value

    def close(self):
        if self.fhandles:
            for f in self.fhandles:
                f.close()

    @classmethod
    def wrapped(cls, dryrun, data, json, files, filepaths):
        """
        Args:
            dryrun (bool | UNSET | None): Optionally override current dryrun setting
            data (dict | tuple | bytes | file | None): Data to send in the body
            json: (optional) json to send in the body
            files (dict | None): File-like-objects for multipart encoding upload.
            filepaths (dict[str, Path] | None): File-like-objects for multipart encoding upload.

        Returns:
            (DataState | None): State, if needed
        """
        if data or json or files or filepaths:
            state = DataState()
            if isinstance(data, Path):
                data = state.add_fh(data, dryrun)

            state.data = data
            state.json = json
            if filepaths:
                files = {}  # Convenience / simplified multipart upload
                for k, fpath in filepaths.items():
                    files[k] = (fpath.name, state.add_fh(fpath, dryrun))

            state.files = files
            return state


class MockResponse:

    def __init__(self, status_code, content):
        self.status_code = status_code
        if content is not None and not isinstance(content, bytes):
            if not isinstance(content, str):
                content = json.dumps(content)

            content = content.encode("utf-8")

        self.content = content

    def json(self):
        return json.loads(self.text)

    @property
    def text(self):
        return stringified(self.content)


class MockedHandlerStack:

    def __init__(self):
        self.handler = None
        self.ms = None
        self.specs = {}
        self.spec_stack = []
        self.original_send_function = None

    def __repr__(self):
        name = self.handler and self.handler.__name__
        status = "active" if self.original_send_function else "stopped"
        nested = " [depth: %s]" % len(self.spec_stack) if self.spec_stack else ""
        return "%s mock %s, %s specs%s" % (name, status, len(self.specs), nested)

    def register_handler(self, handler):
        if self.handler:
            assert self.handler is handler, "Mocks targeting multiple handlers is not supported"
            return

        self.ms = handler.intercept(None)
        self.handler = handler

    def _intercept(self, *args, **kwargs):
        return self.handler.intercept(self, *args, **kwargs)

    def start(self, specs):
        if not self.spec_stack:
            assert self.original_send_function is None
            target, name = self.ms
            self.original_send_function = getattr(target, name)
            setattr(target, name, self._intercept)

        self.spec_stack.append(dict(self.specs))  # Push snapshot of current specs, to be restored by .stop()
        self.specs.update(specs)

    def stop(self):
        self.specs = self.spec_stack.pop()
        if not self.spec_stack:
            assert self.original_send_function is not None
            target, name = self.ms
            setattr(target, name, self.original_send_function)
            self.original_send_function = None

    def response_for_url(self, method, url):
        spec = self.specs.get(url)
        if spec is None:
            return MockResponse(404, dict(message="Default status code 404"))

        if isinstance(spec, BaseException):
            raise spec

        if isinstance(spec, type) and issubclass(spec, BaseException):
            raise spec("Simulated crash")

        if callable(spec):
            spec = spec(method, url)

        if isinstance(spec, MockResponse):
            return spec

        if isinstance(spec, int):
            return MockResponse(spec, "status %s" % spec)

        if isinstance(spec, tuple) and len(spec) == 2:
            return MockResponse(*spec)

        return MockResponse(200, spec)


class MockCentral:

    _stacks = {}

    @classmethod
    def get_stack(cls, handler, key):
        """Keep mock specifications stacked by decorated function, to allow stacking several mock specs per function"""
        stack = cls._stacks.get(key)
        if stack is None:
            stack = MockedHandlerStack()
            cls._stacks[key] = stack

        stack.register_handler(handler)
        return stack


class MockWrapper:
    """Intercept and mock out outgoing RestClient calls"""

    def __init__(self, handler, base_url, specs):
        self.handler = handler
        self.specs = specs or {}
        self.key = None
        self.stack = None
        if base_url:
            assert isinstance(base_url, str)
            assert isinstance(self.specs, dict)
            self.specs = {urljoin(base_url, k): v for k, v in self.specs.items()}

    def __repr__(self):
        status = "stopped" if self.stack is None else "started"
        return "%s %s, %s specs" % (self.key, status, len(self.specs))

    def start(self):
        self.stack = MockCentral.get_stack(self.handler, self.key)
        self.stack.start(self.specs)

    def stop(self):
        self.stack.stop()
        self.stack = None

    def __call__(self, func):
        """
        Args:
            func (callable): We're used as a decorator of function 'func'

        Returns:
            (callable): Decorated function
        """
        self.key = "%s.%s" % (func.__module__, func.__qualname__)

        @functools.wraps(func)
        def inner(*args, **kwargs):
            self.start()
            try:
                return func(*args, **kwargs)

            finally:
                self.stop()

        return inner

    def __enter__(self):
        """We're used as a context"""
        self.key = str(find_caller())
        self.start()
        return self

    def __exit__(self, *_):
        self.stop()


class RestResponse:
    """Simplified response, from a typical REST query, vaguely similar to requests.Response"""

    def __init__(self, method, url, raw_response):
        """
        Args:
            method (str): Method used to query url
            url (url): Remote URL that was queried
            raw_response: Should have similar API to requests.Response
        """
        self.method = method
        self.url = url
        self.raw_response = raw_response
        self.status_code = raw_response.status_code

    def __repr__(self):
        return "<Response [%s]>" % self.status_code

    def json(self):
        return self.raw_response.json()

    @property
    def content(self):
        return self.raw_response.content

    @property
    def ok(self):
        return self.status_code and self.status_code < 400

    @property
    def text(self):
        return self.raw_response.text

    def description(self, size=1024):
        """Description showing what happened with requests 'response', fir .debug() output, or .error() reporting"""
        msg = "%s %s [%s]" % (self.method, self.url, self.status_code)
        if not self.ok and self.status_code != 404:
            msg += " %s" % self.error_reason()

        if len(msg) > size:
            msg = "%s..." % msg[:size]

        return msg

    def error_reason(self):
        """Meaningful error reason from 'response'"""
        try:
            msg = self.extract_message(self.json())
            if msg:
                return msg

        except Exception:  # nosec
            pass

        return self.text

    @staticmethod
    def extract_message(data):
        """Try to get to the point of a returned REST json bit, extract meaningful message if possible"""
        if not data:
            return None

        if data and isinstance(data, str):
            return data.strip()

        if isinstance(data, dict):
            msg = RestResponse.extract_message(data.get("message"))
            if msg:
                return msg

            msg = RestResponse.extract_message(data.get("error"))
            if msg:
                return msg

            return RestResponse.extract_message(data.get("errors"))

        if isinstance(data, list):
            for item in data:
                msg = RestResponse.extract_message(item)
                if msg:
                    return msg


class RestHandler:
    """Allows to use multiple http(s) implementations"""

    @classmethod
    def mock(cls, base_url, specs=None):
        """
        Usage example:

        MY_CLIENT = RestClient("https://example.com")

        @MY_CLIENT.mock({
            "foo": {"some": "payload"}
        })
        def test_foo():
            assert MY_CLIENT.get("foo") == {"some": "payload"}

        Args:
            base (str | None): Base url (all urls in given 'spec' are relative to the base url)
            specs (dict): Map of relative url -> what to return

        Returns:
            Function decorator that will enact the mock
        """
        if callable(base_url):
            # We were invoked without arguments, form: @RestHandler.mock
            w = MockWrapper(cls, None, None)
            return w(base_url)

        return MockWrapper(cls, base_url, specs)

    @classmethod
    def new_session(cls, **session_spec):
        """New session for 'session_spec'"""

    @classmethod
    def raw_response(cls, session, method, url, **passed_through):
        """
        Args:
            session: Session as obtained via new_session() call from this handler
            method (str): Underlying method to call (GET, PUT, POST, etc)
            url (str): Absolute remote URL
            **passed_through: Passed through to underlying call
        """

    @classmethod
    def to_rest_response(cls, method, url, raw_response):
        """
        Args:
            method (str): Underlying method to call (GET, PUT, POST, etc)
            url (str): Absolute remote URL
            raw_response: Raw response as received by underlying call

        Returns:
            (RestResponse): Our simplified response object
        """

    @classmethod
    def intercept(cls, mock_caller, *_, **__):
        """
        Args:
            mock_caller (MockedHandlerStack | None): If provided: effectively intercept, if None: return target+name of function to replace
        """

    @classmethod
    def user_agent(cls):
        """Default user agent to use"""
        return "%s/%s (%s)" % (SYS_INFO.program_name, SYS_INFO.program_version, SYS_INFO.platform_info)


class RequestsHandler(RestHandler):
    """Using requests (client is to bring in the dependency)"""

    @classmethod
    def default_retry(cls):
        import urllib3

        return urllib3.Retry(backoff_factor=1, status_forcelist={413, 429, 500, 502, 503, 504})

    @classmethod
    def get_adapter(cls, retry=None):
        from requests.adapters import HTTPAdapter

        return HTTPAdapter(max_retries=retry or cls.default_retry())

    @classmethod
    def new_session(cls, http_adapter=None, https_adapter=None, retry=None):
        import requests

        session = requests.Session()
        if retry is None:
            retry = cls.default_retry()

        if retry and https_adapter is None:
            https_adapter = cls.get_adapter(retry)

        if retry and http_adapter is None:
            http_adapter = cls.get_adapter(retry)

        if https_adapter:
            session.mount("https://", https_adapter)

        if http_adapter:
            session.mount("http://", http_adapter)

        return session

    @classmethod
    def raw_response(cls, session, method, url, **passed_through):
        func = getattr(session, method.lower(), None)
        if func is None:
            return session.request(method, url, **passed_through)

        return func(url, **passed_through)

    @classmethod
    def to_rest_response(cls, method, url, raw_response):
        return RestResponse(method, url, raw_response)

    @classmethod
    def intercept(cls, mock_caller, *args, **__):
        if mock_caller is None:
            from requests.adapters import HTTPAdapter

            return HTTPAdapter, "send"

        from requests import Response

        request = args[0]
        mocked = mock_caller.response_for_url(request.method, request.url)
        r = Response()
        r.encoding = "utf-8"
        r.status_code = mocked.status_code
        r.url = request.url
        r.request = request
        r._content = mocked.content
        return r

    @classmethod
    def user_agent(cls):
        import requests

        return "%s requests/%s" % (super().user_agent(), requests.__version__)


class RestClient:
    """REST client with good defaults for retry, timeout, ... + support for --dryrun mode etc"""

    handler = RequestsHandler

    def __init__(self, base_url=None, headers=None, timeout=30, cache=None, user_agent=None, handler=None, session=None, **session_spec):
        """
        Args:
            base_url (str | None): Base url of remote REST server
            headers (dict | None): Default headers to use
            timeout (int): Default timeout in seconds
            cache (CacheWrapper): Optional local cache to use
            user_agent (str | None): User-Agent to use for outgoing calls coming from this client (default: handler.user_agent())
            handler: Optional: override default handler
            session: Optional: override getting handler.new_session()
        """
        self.base_url = base_url
        self.headers = dict(headers) if headers else {}
        self.timeout = timeout
        self.cache_wrapper = cache
        if handler:
            self.handler = handler

        self.user_agent = user_agent or self.handler.user_agent()
        self.session = session or self.handler.new_session(**session_spec)
        if self.user_agent:
            self.headers["User-Agent"] = self.user_agent

    def __repr__(self):
        return stringified(self.base_url or self.session)

    def sub_client(self, relative_url):
        """
        Args:
            relative_url (str): Relative url (relative to self.base_url)

        Returns:
            (RestClient): Same as current client, with a different/child base url
        """
        url = urljoin(self.base_url, relative_url)
        return RestClient(
            url, headers=self.headers, timeout=self.timeout, user_agent=self.user_agent, handler=self.handler, session=self.session
        )

    def full_url(self, url):
        """
        Args:
            url (str): Relative URL

        Returns:
            (str): Absolute URL, auto-completed with `self.base_url`
        """
        return urljoin(self.base_url, url)

    def decompress(self, url, destination, simplify=False, fatal=True, logger=UNSET, dryrun=UNSET, **kwargs):
        """
        Args:
            url (str): URL of .tar.gz to unpack (may be absolute, or relative to self.base_url)
            destination (str | Path): Path to local folder where to untar url
            simplify (bool): If True and source has only one sub-folder, extract that one sub-folder to destination
            fatal (type | bool | None): True: abort execution on failure, False: don't abort but log, None: don't abort, don't log
            logger (callable | bool | None): Logger to use, True to print(), False to trace(), None to disable log chatter
            dryrun (bool | UNSET | None): Optionally override current dryrun setting
            **kwargs: Passed through to underlying client

        Returns:
            (RestResponse): Response from underlying call
        """
        _, _, actual_url = self._decomposed_checksum_url(url)
        destination = to_path(destination).absolute()
        with TempFolder():
            tarball_path = to_path(os.path.basename(actual_url)).absolute()
            response = self.download(url, tarball_path, fatal=fatal, logger=logger, dryrun=dryrun, **kwargs)
            if response.ok:
                decompress(tarball_path, destination, simplify=simplify, fatal=fatal, logger=logger, dryrun=dryrun)

            return response

    def download(self, url, destination, fatal=True, logger=UNSET, dryrun=UNSET, **kwargs):
        """
        Args:
            url (str): URL of resource to download (may be absolute, or relative to self.base_url)
                       Use #sha256=... or #sha512=... at the end of the url to ensure content is validated against given checksum
            destination (str | Path): Path to local file where to store the download
            fatal (type | bool | None): True: abort execution on failure, False: don't abort but log, None: don't abort, don't log
            logger (callable | bool | None): Logger to use, True to print(), False to trace(), None to disable log chatter
            dryrun (bool | UNSET | None): Optionally override current dryrun setting
            **kwargs: Passed through to underlying client

        Returns:
            (RestResponse): Response from underlying call
        """
        hash_algo, hash_checksum, url = self._decomposed_checksum_url(url)
        response = self._get_response("GET", url, fatal, logger, dryrun=dryrun, action="download", **kwargs)
        if response.ok and not _R.resolved_dryrun(dryrun):
            destination = to_path(destination)
            ensure_folder(destination.parent, fatal=fatal, logger=None)
            with open(destination, "wb") as fh:
                fh.write(response.content)

            if hash_checksum:
                downloaded_checksum = checksum(destination, hash=hash_algo)
                if downloaded_checksum != hash_checksum:
                    delete(destination, fatal=False, logger=logger)
                    msg = "%s differs for %s: expecting %s, got %s" % (hash_algo, short(destination), hash_checksum, downloaded_checksum)
                    return abort(msg, fatal=fatal, logger=logger)

        return response

    def get_response(self, url, fatal=False, logger=False, **kwargs):
        """
        Args:
            url (str): Remote URL (may be absolute, or relative to self.base_url)
            fatal (type | bool | None): True: abort execution on failure, False: don't abort but log, None: don't abort, don't log
            logger (callable | bool | None): Logger to use, True to print(), False to trace(), None to disable log chatter
            **kwargs: Passed through to underlying client

        Returns:
            (RestResponse): Response from underlying call
        """
        return self._get_response("GET", url, fatal, logger, **kwargs)

    def delete(self, url, fatal=True, logger=UNSET, dryrun=UNSET, **kwargs):
        """Same as underlying .delete(), but respecting 'dryrun' mode

        Args:
            url (str): URL to query (can be absolute, or relative to self.base_url)
            fatal (type | bool | None): True: abort execution on failure, False: don't abort but log, None: don't abort, don't log
            logger (callable | bool | None): Logger to use, True to print(), False to trace(), None to disable log chatter
            dryrun (bool | UNSET | None): Optionally override current dryrun setting
            **kwargs: Passed through to underlying client

        Returns:
            (RestResponse): Response from underlying call
        """
        return self._get_response("DELETE", url, fatal, logger, dryrun=dryrun, **kwargs)

    def get(self, url, fatal=False, logger=False, **kwargs):
        """
        Args:
            url (str): Remote URL (may be absolute, or relative to self.base_url)
            fatal (type | bool | None): True: abort execution on failure, False: don't abort but log, None: don't abort, don't log
            logger (callable | bool | None): Logger to use, True to print(), False to trace(), None to disable log chatter
            **kwargs: Passed through to underlying client

        Returns:
            (dict | None): Deserialized .json() from response, if available
        """
        response = self.get_response(url, fatal=fatal, logger=logger, **kwargs)
        if response.ok:
            return response.json()

    def head(self, url, fatal=False, logger=False, **kwargs):
        """
        Args:
            url (str): URL to query (can be absolute, or relative to self.base_url)
            fatal (type | bool | None): True: abort execution on failure, False: don't abort but log, None: don't abort, don't log
            logger (callable | bool | None): Logger to use, True to print(), False to trace(), None to disable log chatter
            **kwargs: Passed through to underlying client

        Returns:
            (RestResponse): Response from underlying call
        """
        return self._get_response("HEAD", url, fatal, logger, **kwargs)

    def post(self, url, fatal=True, logger=UNSET, dryrun=UNSET, data=None, json=None, files=None, filepaths=None, **kwargs):
        """Same as underlying .post(), but respecting 'dryrun' mode

        Args:
            url (str): URL to query (can be absolute, or relative to self.base_url)
            fatal (type | bool | None): True: abort execution on failure, False: don't abort but log, None: don't abort, don't log
            logger (callable | bool | None): Logger to use, True to print(), False to trace(), None to disable log chatter
            dryrun (bool | UNSET | None): Optionally override current dryrun setting
            data (dict | tuple | bytes | file | None): Data to send in the body
            json: (optional) json to send in the body
            files (dict | None): File-like-objects for multipart encoding upload.
            filepaths (dict[str, Path] | None): File-like-objects for multipart encoding upload.
            **kwargs: Passed through to underlying client

        Returns:
            (RestResponse): Response from underlying call
        """
        state = DataState.wrapped(dryrun, data, json, files, filepaths)
        return self._get_response("POST", url, fatal, logger, dryrun=dryrun, state=state, **kwargs)

    def purge(self, url, fatal=True, logger=UNSET, dryrun=UNSET, **kwargs):
        """Same as underlying .purge(), but respecting 'dryrun' mode

        Args:
            url (str): URL to query (can be absolute, or relative to self.base_url)
            fatal (type | bool | None): True: abort execution on failure, False: don't abort but log, None: don't abort, don't log
            logger (callable | bool | None): Logger to use, True to print(), False to trace(), None to disable log chatter
            dryrun (bool | UNSET | None): Optionally override current dryrun setting
            **kwargs: Passed through to underlying client

        Returns:
            (RestResponse): Response from underlying call
        """
        return self._get_response("PURGE", url, fatal, logger, dryrun=dryrun, **kwargs)

    def put(self, url, fatal=True, logger=UNSET, dryrun=UNSET, data=None, json=None, files=None, filepaths=None, **kwargs):
        """
        Args:
            url (str): URL to query (can be absolute, or relative to self.base_url)
            fatal (type | bool | None): True: abort execution on failure, False: don't abort but log, None: don't abort, don't log
            logger (callable | bool | None): Logger to use, True to print(), False to trace(), None to disable log chatter
            dryrun (bool | UNSET | None): Optionally override current dryrun setting
            data (dict | tuple | bytes | file | None): Data to send in the body
            json: (optional) json to send in the body
            files (dict | None): File-like-objects for multipart encoding upload.
            filepaths (dict[str, Path] | None): File-like-objects for multipart encoding upload.
            **kwargs: Passed through to underlying client

        Returns:
            (RestResponse): Response from underlying call
        """
        state = DataState.wrapped(dryrun, data, json, files, filepaths)
        return self._get_response("PUT", url, fatal, logger, dryrun=dryrun, state=state, **kwargs)

    def url_exists(self, url, logger=False, **kwargs):
        """
        Args:
            url (str): URL to query (can be absolute, or relative to self.base_url)
            logger (callable | bool | None): Logger to use, True to print(), False to trace(), None to disable log chatter
            **kwargs: Passed through to underlying client

        Returns:
            (bool): True if remote URL exists (ie: not a 404)
        """
        response = self.head(url, logger=logger, **kwargs)
        return bool(response and response.ok)

    def mock(self, specs):
        """
        Usage example:

        MY_CLIENT = RestClient("https://example.com")

        @MY_CLIENT.mock({
            "foo": {"some": "payload"}
        })
        def test_foo():
            assert MY_CLIENT.get("foo") == {"some": "payload"}

        Args:
            specs (dict): Map of relative url -> what to return

        Returns:
            Function decorator that will enact the mock
        """
        if callable(specs):
            # We were invoked without arguments, form: @MY_CLIENT.mock
            w = MockWrapper(self.handler, self.base_url, None)
            return w(specs)

        return MockWrapper(self.handler, self.base_url, specs)

    @staticmethod
    def std_diskcache(directory=UNSET, default_expire=UNSET, size_limit=UNSET):
        """
        Convenience method to obtain a diskcache with good defaults, callers must require diskcache (not provided by runez)
        This can be used as an example cache setup

        Args:
            directory (str | Path | None): Directory where to store the cache
            default_expire (int | float | str): Default expiration time in seconds
            size_limit (int | float | str): Size limit for this cache

        Returns:
            (CacheWrapper): Object wrapping this cache
        """
        try:
            from diskcache import Cache

            if directory is UNSET:
                directory = None
                if not DEV.current_test():
                    # By default, do not use ~/.cache for test runs (diskcache will default to a temp dir)
                    pymm = "py%s" % joined(sys.version_info[:2], delimiter="")
                    directory = CacheWrapper.cache_base_path(suffix=pymm)

            if default_expire is UNSET:
                default_expire = CacheWrapper.default_expire

            if size_limit is UNSET:
                size_limit = _R.lc.rm.to_bytesize(CacheWrapper.size_limit)

            cache_backend = Cache(directory=directory or None, size_limit=size_limit)
            return CacheWrapper(cache_backend, directory, default_expire,  size_limit)

        except ImportError:
            return None

    def _protected_get(self, method, absolute_url, keyword_args):
        try:
            return self.handler.raw_response(self.session, method, absolute_url, **keyword_args)

        except ForbiddenHttpError:
            pass  # Shorten stack trace

        raise ForbiddenHttpError(absolute_url)

    @classmethod
    def _decomposed_checksum_url(cls, url):
        regex = getattr(cls, "_checksum_regex", None)
        if regex is None:
            regex = cls._checksum_regex = re.compile(r"#(md5|sha(1|256|512))=([a-f0-9]+)")

        m = regex.search(url)
        if m and m.end(0) == len(url):
            hash_algo = m.group(1)
            hash_checksum = m.group(3)
            return hash_algo, hash_checksum, url[:m.start(0)]

        return None, None, url

    def _get_response(self, method, url, fatal, logger, dryrun=False, state=None, action=None, **kwargs):
        """
        Args:
            method (str): Underlying method to call
            url (str): Remote URL (may be absolute, or relative to self.base_url)
            fatal (type | bool | None): True: abort execution on failure, False: don't abort but log, None: don't abort, don't log
            logger (callable | bool | None): Logger to use, True to print(), False to trace(), None to disable log chatter
            dryrun (bool | UNSET | None): Optionally override current dryrun setting
            state (DataState | None): For PUT/POST requests
            action (str | None): Action to refer to in dryrun message (default: method)
            **kwargs: Passed through to underlying client

        Returns:
            (RestResponse): Response from underlying call
        """
        absolute_url = self.full_url(url)
        message = "%s %s" % (action or method, absolute_url)
        if _R.hdry(dryrun, logger, message):
            return RestResponse(method, absolute_url, MockResponse(200, dict(message="dryrun %s" % message)))

        cache_key = cache_expire = None
        if self.cache_wrapper is not None:
            cache_expire = kwargs.pop("expire", UNSET)
            cache_key = self.cache_wrapper.cache_key(absolute_url, params=kwargs.get("params"))
            if method in ("PURGE", "DELETE"):
                self.cache_wrapper.delete(cache_key)

            elif self.cache_wrapper.is_cachable_method(method):
                response = self.cache_wrapper.get(cache_key)
                if response is not None:
                    return response

        full_headers = self.headers
        headers = kwargs.get("headers")
        if headers:
            full_headers = dict(full_headers)
            full_headers.update(headers)

        keyword_args = dict(kwargs)
        keyword_args["headers"] = full_headers
        keyword_args.setdefault("timeout", self.timeout)
        if state is not None:
            state.complete(keyword_args)

        try:
            raw_response = self._protected_get(method, absolute_url, keyword_args)
            response = self.handler.to_rest_response(method, absolute_url, raw_response)
            if fatal or logger is not None:
                msg = response.description()
                if fatal and not response.ok:
                    abort(msg, fatal=fatal, logger=logger)

                _R.hlog(logger, msg)

            if cache_key is not None and response.ok and self.cache_wrapper.is_cachable_method(method):
                self.cache_wrapper.set(cache_key, response, expire=cache_expire)

            return response

        finally:
            if state is not None:
                state.close()



import math

import runez


def test_boolean():
    assert runez.to_boolean(None) is False
    assert runez.to_boolean("") is False
    assert runez.to_boolean("t") is False
    assert runez.to_boolean("0") is False
    assert runez.to_boolean("0.0") is False
    assert runez.to_boolean("1.0.0") is False

    assert runez.to_boolean("True") is True
    assert runez.to_boolean("Y") is True
    assert runez.to_boolean("yes") is True
    assert runez.to_boolean("On") is True
    assert runez.to_boolean("5") is True
    assert runez.to_boolean("0.1") is True
    assert runez.to_boolean("16.1") is True


def test_bytesize():
    assert runez.to_bytesize(10) == 10
    assert runez.to_bytesize(None) is None
    assert runez.to_bytesize("") is None
    assert runez.to_bytesize("1a") is None

    assert runez.to_bytesize("10") == 10
    assert runez.to_bytesize("10.4") == 10
    assert runez.to_bytesize("10.6") == 11
    assert runez.to_bytesize(10, default_unit="k", base=1024) == 10 * 1024
    assert runez.to_bytesize(10, default_unit="k", base=1000) == 10000
    assert runez.to_bytesize("10", default_unit="k", base=1000) == 10000
    assert runez.to_bytesize("10m", default_unit="k", base=1000) == 10000000
    assert runez.to_bytesize("10.m", default_unit="k", base=1000) == 10000000
    assert runez.to_bytesize("10.4m", default_unit="k", base=1000) == 10400000

    assert runez.to_bytesize(10, default_unit="a", base=1000) is None  # Bogus default_unit


def test_plural():
    assert runez.plural(None) == "no values"
    assert runez.plural(None, None) == "no values"
    assert runez.plural("chair") == "chairs"
    assert runez.plural(None, "walker") == "no walkers"

    assert runez.plural(None) == "no values"
    assert runez.plural(1) == "1 value"
    assert runez.plural("foo", "character") == "3 characters"

    assert runez.plural(2, "match") == "2 matches"
    assert runez.plural(0, "dish") == "0 dishes"
    assert runez.plural(5, "goth") == "5 goths"
    assert runez.plural(7, "diff") == "7 diffs"

    assert runez.plural(1, "penny") == "1 penny"
    assert runez.plural(2, "penny") == "2 pennies"

    assert runez.plural([], "record") == "0 records"
    assert runez.plural([""], "record") == "1 record"

    assert runez.plural(1, "person") == "1 person"
    assert runez.plural(2, "person") == "2 people"

    assert runez.plural(2, "man") == "2 men"
    assert runez.plural(2, "woman") == "2 women"
    assert runez.plural(2, "status") == "2 statuses"

    assert runez.plural(0, "carrot") == "0 carrots"
    assert runez.plural(1, "carrot") == "1 carrot"
    assert runez.plural(20, "carrot") == "20 carrots"
    assert runez.plural(20000, "carrot") == "20K carrots"
    assert runez.plural(20000, "carrot", base=None) == "20K carrots"
    assert runez.plural(20000, "carrot", base=0) == "20000 carrots"


def test_representation():
    assert runez.represented_bytesize(20) == "20 B"
    assert runez.represented_bytesize(20, unit="") == "20"
    assert runez.represented_bytesize(9000) == "8.8 KB"
    assert runez.represented_bytesize(20000) == "20 KB"
    assert runez.represented_bytesize(20000, unit="") == "20 K"
    assert runez.represented_bytesize(20000000) == "19 MB"
    assert runez.represented_bytesize(20000000000) == "19 GB"
    assert runez.represented_bytesize(20000000000000) == "18 TB"
    assert runez.represented_bytesize(20000000000000000) == "18 PB"
    assert runez.represented_bytesize(20000000000000000000) == "17764 PB"

    assert runez.represented_with_units(20) == "20"
    assert runez.represented_with_units(999.9) == "999.9"
    assert runez.represented_with_units(1001) == "1K"
    assert runez.represented_with_units(1060) == "1.1K"
    assert runez.represented_with_units(8900) == "8.9K"
    assert runez.represented_with_units(20304050600000000000) == "20304P"


SAMPLE_TABULATED_OUTPUT = """
  UID   PID  PPID   C STIME   TTY           TIME CMD
  501 96590    42   0 foo bar ttys009    0:04.60 -bash
  502     4     5   0         ttys009            zsh
--
  UID   PID  PPID   C STIME   TTY           TIME CMD
  501 96590    42   0 18Dec20 ttys009    0:04.60 -bash --login
--
UID        PID  PPID  C STIME TTY      STAT   TIME CMD
zoran        7    42  0 15:23 tty1     S      0:00 /bin/bash --login
--
"""


def tabulated_samples():
    lines = []
    for line in SAMPLE_TABULATED_OUTPUT.splitlines():
        if not line or line.startswith("#"):
            continue

        if line.startswith("--"):
            yield len(lines) - 1, "\n".join(lines)
            lines = []
            continue

        lines.append(line)

    if lines:
        yield len(lines) - 1, "\n".join(lines)


def test_tabulated_parsing():
    assert runez.parsed_tabular("  \nfoo") == []  # First line must have a header...

    for expected, output in tabulated_samples():
        parsed = list(runez.parsed_tabular(output))
        assert len(parsed) == expected
        assert "bash" in parsed[0]["CMD"]


DOCKER_PS_SAMPLE = """
CONTAINER_ID  IMAGE                  COMMAND  CREATED         STATUS         PORTS                   NAMES
d7907cca5247  foo/bar:stable         "/init"  18 minutes ago  Up 18 minutes                          foo
6bda1a6f83eb  linuxserver/syncthing  "/init"  7 days ago      Up 43 hours    0.0.0.0:123->123/tcp    syncthing
"""


def test_tabulated_docker_ps():
    info = runez.parsed_tabular(DOCKER_PS_SAMPLE.strip())
    assert info == [
        {
            "CONTAINER_ID": "d7907cca5247",
            "IMAGE": "foo/bar:stable",
            "COMMAND": '"/init"',
            "CREATED": "18 minutes ago",
            "STATUS": "Up 18 minutes",
            "NAMES": "foo"
        },
        {
            "CONTAINER_ID": "6bda1a6f83eb",
            "IMAGE": "linuxserver/syncthing",
            "COMMAND": '"/init"',
            "CREATED": "7 days ago",
            "STATUS": "Up 43 hours",
            "PORTS": "0.0.0.0:123->123/tcp",
            "NAMES": "syncthing"
        }
    ]


def test_to_float():
    assert runez.to_float(None) is None
    assert runez.to_float("foo") is None
    assert runez.to_float("-.if") is None
    assert runez.to_float(["foo"], lenient=True) is None

    assert runez.to_float("0") == 0.0
    assert runez.to_float("-5") == -5.0
    assert runez.to_float("15") == 15.0
    assert runez.to_float("+135.057E+4") == 1350570.0
    assert runez.to_float("-135000.5e-3") == -135.0005

    assert math.isnan(runez.to_float("nan"))
    assert math.isinf(runez.to_float("inf"))
    assert math.isinf(runez.to_float(".inf"))
    assert math.isinf(runez.to_float("-.inf"))

    assert isinstance(runez.to_float("15"), float)
    assert isinstance(runez.to_float("15", lenient=True), int)


def test_to_int():
    assert runez.to_int(None) is None
    assert runez.to_int("foo") is None
    assert runez.to_int(["foo"]) is None
    assert runez.to_int("5.0") is None
    assert runez.to_int("_5_") is None
    assert runez.to_int("_5") is None
    assert runez.to_int("5_") is None
    assert runez.to_int("1__5") is None
    assert runez.to_int("1_ 5") is None

    assert runez.to_int("0") == 0
    assert runez.to_int("-5") == -5
    assert runez.to_int("15") == 15

    assert runez.to_int("0o10") == 8
    assert runez.to_int("0x10") == 16
    assert runez.to_int("0o1_0") == 8
    assert runez.to_int("0x1_0") == 16

    assert runez.to_int(" 1_500 ") == 1500
    assert runez.to_int("1_5 ") == 15
    assert runez.to_int("1_500_001 ") == 1500001


def test_wordification():
    assert runez.identifiers(None) == []
    assert runez.identifiers("") == []
    assert runez.identifiers("a_b1") == ["a_b1"]
    assert runez.identifiers("hi_There-you") == ["hi_There", "you"]
    assert runez.identifiers(["a", ["b_c(d)", None, [1]]]) == ["a", "b_c", "d", "1"]
    assert runez.identifiers({"a": "b_c(d)"}) == ["a", "b_c", "d"]

    assert runez.words(None) == []
    assert runez.words("") == []
    assert runez.words("*") == []
    assert runez.words("a") == ["a"]
    assert runez.words("a b") == ["a", "b"]
    assert runez.words("a,b(c)") == ["a", "b", "c"]
    assert runez.words("a,,b") == ["a", "b"]
    assert runez.words("a_b1") == ["a", "b1"]
    assert runez.words("hi_There-you", normalize=str.lower) == ["hi", "there", "you"]
    assert runez.words({"a": "b_c(d)"}) == ["a", "b", "c", "d"]

    # Test de-camelization
    assert runez.words("foo", decamel=True) == ["foo"]
    assert runez.words("fooBar", decamel=True) == ["foo", "Bar"]
    assert runez.words("FooBar", decamel=True) == ["Foo", "Bar"]
    assert runez.words("_FooBar_", decamel=True, normalize=str.lower) == ["foo", "bar"]
    assert runez.words("someCamel_likeWords-FooBar_baz", decamel=True) == ['some', 'Camel', 'like', 'Words', 'Foo', 'Bar', 'baz']

    assert runez.words(["a", "b_c", "a"]) == ["a", "b", "c", "a"]
    assert runez.words(["a", None, "b", 1]) == ["a", "b", "1"]
    assert runez.words(["a", [None, "b,c"], 1]) == ["a", "b", "c", "1"]

    assert runez.wordified(None) is None
    assert runez.wordified("Hello_There", delimiter="-") == "Hello-There"

    assert runez.snakified("my-key") == "MY_KEY"
    assert runez.camel_cased("my-key") == "MyKey"
    assert runez.entitled("my-key") == "My key"

    assert runez.affixed(None) is None
    assert runez.affixed("") == ""
    assert runez.affixed("", prefix="my-") == "my-"

    assert runez.affixed("my-key") == "my-key"
    assert runez.affixed("my-key", prefix="my-") == "my-key"
    assert runez.affixed("key", prefix="my-") == "my-key"
    assert runez.affixed("my-key", prefix="X_", normalize=runez.snakified) == "X_MY_KEY"
    assert runez.affixed("my-key", prefix="X_", suffix="+foo") == "X_my-key+foo"



import threading

from runez.system import py_mimic


THREAD_LOCAL = threading.local()


class thread_local_property:
    """
    A property that is computed once per thread
    Use this in rare cases where you need just a property (or 2) to be thread local in a given object

    If you need an entire object to be thread-local, then use class `ThreadLocalSingleton` instead
    """

    def __init__(self, func):
        self.__func__ = func
        self.key = "_tp%s" % func.__qualname__
        py_mimic(self, func)

    def __get__(self, instance, owner):
        if instance is None:
            return self

        if not hasattr(THREAD_LOCAL, self.key):
            setattr(THREAD_LOCAL, self.key, self.__func__(instance))

        return getattr(THREAD_LOCAL, self.key)


class ThreadLocalSingleton:
    """
    Class ancestor intended to easily allow you to get per-thread singletons

    Usage:
        class MyClass(ThreadLocalSingleton):
            def __init__(...):
                # Will be called once per thread, no matter how many times or where it gets invoked from
                ...
    """

    def __new__(cls, *positional, **named):
        # We could do singleton by combination of args, but outcome of that could be hard to grok for users
        # Not sure if there's a good use case for this (one where gotcha-factor is much lower than added value)
        assert not positional and not named, "Current limitation: only classes that can be created without args are supported for now"

        key = "_ts%s.%s" % (cls.__module__, cls.__name__)
        existing = getattr(THREAD_LOCAL, key, None)
        if existing is None:
            existing = super().__new__(cls)
            setattr(THREAD_LOCAL, key, existing)

        return existing
